from .protocols import MessageObject, Codec
from .base import BaseMessageObject, BaseCodec
from .recording import (get_recording, get_recording_from_file, get_recording_codec, BufferObject, BufferCodec,
                        recorded_packet)
from .contrib import JSONCodec, JSONObject




import json
from dataclasses import dataclass

from aionetworking.formats.base import BaseCodec, BaseMessageObject

from typing import Any, AsyncGenerator, Tuple


@dataclass
class JSONCodec(BaseCodec):
    codec_name = 'json'

    """
    Decode & Encode JSON text messages
    """

    async def decode(self, encoded: bytes, **kwargs) -> AsyncGenerator[Tuple[bytes, Any], None]:
        pos = 0
        end = len(encoded)
        while pos < end:
            start = pos
            data = encoded.decode()
            msg, pos = json.JSONDecoder().raw_decode(data, idx=pos)
            yield encoded[start:pos], msg

    async def encode(self, decoded: Any, **kwargs) -> bytes:
        return json.dumps(decoded).encode()


@dataclass
class JSONObject(BaseMessageObject):
    name = 'JSON'
    codec_cls = JSONCodec

    def __getattr__(self, item):
        if item in self.decoded:
            return self.decoded[item]
        raise AttributeError




import yaml
from .ssl import ServerSideSSL, ClientSideSSL
from .protocol_factories import (StreamServerProtocolFactory, DatagramServerProtocolFactory,
                                 StreamClientProtocolFactory, DatagramClientProtocolFactory)


def ssl_server_side_constructor(loader, node) -> ServerSideSSL:
    value = loader.construct_mapping(node) if node.value else {}
    return ServerSideSSL(**value)


def ssl_client_side_constructor(loader, node) -> ClientSideSSL:
    value = loader.construct_mapping(node) if node.value else {}
    return ClientSideSSL(**value)


def stream_server_protocol_factory_constructor(loader, node) -> StreamServerProtocolFactory:
    value = loader.construct_mapping(node) if node.value else {}
    return StreamServerProtocolFactory(**value)


def datagram_server_protocol_factory_constructor(loader, node) -> DatagramServerProtocolFactory:
    value = loader.construct_mapping(node) if node.value else {}
    return DatagramServerProtocolFactory(**value)


def stream_client_protocol_factory_constructor(loader, node) -> StreamClientProtocolFactory:
    value = loader.construct_mapping(node) if node.value else {}
    return StreamClientProtocolFactory(**value)


def datagram_client_protocol_factory_constructor(loader, node) -> DatagramClientProtocolFactory:
    value = loader.construct_mapping(node) if node.value else {}
    return DatagramClientProtocolFactory(**value)


def load_server_side_ssl(Loader=yaml.SafeLoader):
    yaml.add_constructor('!ServerSideSSL', ssl_server_side_constructor, Loader=Loader)


def load_client_side_ssl(Loader=yaml.SafeLoader):
    yaml.add_constructor('!ClientSideSSL', ssl_client_side_constructor, Loader=Loader)


def load_stream_server_protocol_factory(Loader=yaml.SafeLoader):
    yaml.add_constructor('!StreamServerProtocolFactory', stream_server_protocol_factory_constructor, Loader=Loader)


def load_datagram_server_protocol_factory(Loader=yaml.SafeLoader):
    yaml.add_constructor('!DatagramServerProtocolFactory', datagram_server_protocol_factory_constructor, Loader=Loader)


def load_stream_client_protocol_factory(Loader=yaml.SafeLoader):
    yaml.add_constructor('!StreamClientProtocolFactory', stream_client_protocol_factory_constructor, Loader=Loader)


def load_datagram_client_protocol_factory(Loader=yaml.SafeLoader):
    yaml.add_constructor('!DatagramClientProtocolFactory', datagram_client_protocol_factory_constructor, Loader=Loader)



import asyncssh
from dataclasses import dataclass, InitVar
from pathlib import Path

from aionetworking.utils import IPNetwork, get_ip_port
from aionetworking.networking.sftp import SFTPClientProtocolFactory, SFTPClientProtocol
from .clients import BaseNetworkClient

from typing import Tuple


@dataclass
class SFTPClient(BaseNetworkClient):
    name = "SFTP Client"
    protocol_factory = SFTPClientProtocolFactory
    expected_connection_exceptions = (ConnectionRefusedError, asyncssh.misc.PermissionDenied)
    peer_prefix = 'sftp'
    sftp_log_level: InitVar[int] = 1
    known_hosts: Path = None
    username: str = None
    password: str = None
    client_keys: Path = None
    passphrase: str = None
    client_version: Tuple = ()
    sftp_client = None
    sftp = None
    cwd = None

    def __post_init__(self, sftp_log_level) -> None:
        super().__post_init__()
        asyncssh.logging.set_debug_level(sftp_log_level)

    async def _open_connection(self) -> SFTPClientProtocol:
        self.sftp_conn, self.conn = await asyncssh.create_connection(
            self.protocol_factory, self.host, self.port, local_addr=self.local_addr, known_hosts=self.known_hosts,
            username=self.username, password=self.password, client_keys=self.client_keys, passphrase=self.passphrase,
            client_version = self.client_version)
        """network = IPNetwork(self.host)
        if network.is_ipv6:
            self.actual_srcip, self.actual_srcport, self.flowinfo, self.scope_id = self.sftp_conn.get_extra_info('sockname')
        else:
            self.actual_srcip, self.actual_srcport = self.sftp_conn.get_extra_info('sockname')"""
        await self.conn.wait_context_set()
        self.sftp = await self.sftp_conn.start_sftp_client()
        await self.conn.set_sftp(self.sftp)
        return self.conn

    def is_closing(self) -> bool:
        return self._status.is_stopping_or_stopped()

    async def _close_connection(self) -> None:
        if self.conn and self.sftp:
            await self.conn.wait_tasks_done()
            self.sftp.exit()
            self.conn.close()
            await self.conn.wait_closed()

    @property
    def actual_local_addr(self) -> Tuple[str, int]:
        if self.host and self.sftp_conn:
            self._actual_listening_on = get_ip_port(self.host, self.sftp_conn)
            return self._actual_listening_on
        if self._actual_listening_on:
            return self._actual_listening_on



#!/usr/bin/env python
from pathlib import Path
from aionetworking.receivers.sftp import generate_key_in_path


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument(type=Path, dest='path',
                        help='path to file to store the ssh key')
    parser.add_argument('--algorithm', default='ssh-rsa', type=str,
                        help='algorithim to use to generate the key')
    p = parser.parse_args()
    generate_key_in_path(p.path, p.algorithm)


import asyncio
import os
import sys

py311 = sys.version_info >= (3, 11)
py310 = sys.version_info >= (3, 10)
py39 = sys.version_info >= (3, 9)


def default_server_port() -> int:
    policy = asyncio.get_event_loop_policy()
    base_port = 3110 if py311 else 3100 if py310 else 3900 if py39 else 3800
    if os.name == 'nt':
        if isinstance(policy, asyncio.WindowsProactorEventLoopPolicy):
            return base_port + 10
        if isinstance(policy, asyncio.WindowsSelectorEventLoopPolicy):
            return base_port + 11
        return base_port + 12
    if isinstance(policy, asyncio.DefaultEventLoopPolicy):
        return base_port
    return base_port + 1


def default_client_port() -> int:
    policy = asyncio.get_event_loop_policy()
    base_port = 31100 if py311 else 31000 if py310 else 39000 if py39 else 38000
    if os.name == 'nt':
        if isinstance(policy, asyncio.WindowsProactorEventLoopPolicy):
            return base_port + 100
        if isinstance(policy, asyncio.WindowsSelectorEventLoopPolicy):
            return base_port + 110
        return base_port + 120
    if isinstance(policy, asyncio.DefaultEventLoopPolicy):
        return base_port
    else:
        return base_port + 10




import asyncio
from dataclasses import dataclass, field

from aionetworking.factories import future_defaultdict
from typing import Dict, Sequence, Any, List


class StatusException(BaseException): ...


@dataclass
class ValueWaiter:
    values: Sequence[Any] = ()
    _waiters: Dict[Any, asyncio.Future] = field(default_factory=future_defaultdict, init=False)
    _val: str = field(default='', init=False)
    _historical_values: List[Any] = field(default_factory=list, init=False)

    def has_had_value(self, value: str):
        return value in self._historical_values

    async def wait_has_had_value(self, value: str):
        if self.has_had_value(value):
            return
        if value not in self.values:
            raise StatusException(f'{value} is not a valid value')
        await self._waiters[value]

    def is_value(self, value: str):
        return self._val == value

    def is_one_of(self, values: Sequence[Any]):
        return any(v == self._val for v in values)

    async def wait_value(self, value: str):
        if self._val == value:
            return
        if value not in self.values:
            raise StatusException(f'{value} is not a valid value')
        await self._waiters[value]

    def _check_waiters(self):
        if self._val in self._waiters:
            if not self._waiters[self._val].done():
                self._waiters[self._val].set_result(True)
            self._waiters.pop(self._val)

    def set_value(self, value: str):
        if value not in self.values:
            raise StatusException(f'{value} is not a valid value')
        self._val = value
        if value not in self._historical_values:
            self._historical_values.append(value)
        self._check_waiters()

    def set_exception(self, value: str, exc: BaseException):
        if value in self._waiters:
            self._waiters[value].set_exception(exc)
            self._waiters.pop(value)


@dataclass
class StatusWaiter(ValueWaiter):
    values: Sequence[Any] = ('stopped', 'starting', 'started', 'stopping')
    _val: str = field(default='stopped', init=False)

    async def wait_started(self):
        return await self.wait_value('started')

    async def wait_starting(self):
        return await self.wait_value('starting')

    async def wait_stopped(self):
        return await self.wait_value('stopped')

    async def wait_stopping(self):
        return await self.wait_value('stopping')

    async def wait_starting_or_started(self):
        return await asyncio.wait([self.wait_started(), self.wait_starting()], return_when=asyncio.FIRST_COMPLETED)

    async def wait_stopping_or_stopped(self):
        return await asyncio.wait([self.wait_stopped(), self.wait_stopping()], return_when=asyncio.FIRST_COMPLETED)

    async def wait_has_started(self):
        return await self.wait_has_had_value('started')

    def has_started(self):
        return self.has_had_value('started')

    def is_started(self):
        return self.is_value('started')

    def is_stopped(self):
        return self.is_value('stopped')

    def is_starting_or_started(self):
        return self.is_one_of(['started', 'starting'])

    def is_stopping_or_stopped(self):
        return self.is_one_of(['stopped', 'stopping'])

    def set_started(self):
        self.set_value('started')

    def set_stopped(self):
        self.set_value('stopped')

    def set_stopping(self):
        self.set_value('stopping')

    def set_starting(self):
        self.set_value('starting')



import asyncio
from pathlib import Path
from ssl import SSLContext
import os
import sys

from dataclasses import dataclass, field

from .base import BaseServer, BaseNetworkServer
from aionetworking.networking.connections import UDPServerConnection
from aionetworking.networking.protocol_factories import DatagramServerProtocolFactory, StreamServerProtocolFactory
from aionetworking.networking.ssl import ServerSideSSL
from aionetworking.utils import unix_address, pipe_address
from aionetworking.futures.value_waiters import StatusWaiter

import socket
from typing import List, Optional, Union, Tuple


@dataclass
class TCPServer(BaseNetworkServer):
    name = "TCP Server"
    peer_prefix = 'tcp'
    protocol_factory: StreamServerProtocolFactory = None
    reuse_address: bool = True

    ssl: Optional[ServerSideSSL] = None
    ssl_handshake_timeout: Optional[int] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.ssl:
            self.close_tasks.append(self.ssl.close)
            self.ssl.set_logger(self.logger)

    @property
    def ssl_context(self) -> Optional[SSLContext]:
        if self.ssl:
            return self.ssl.context

    async def _get_server(self) -> asyncio.AbstractServer:
        return await self.loop.create_server(self.protocol_factory, host=self.host, port=self.port,
                                             ssl=self.ssl_context, reuse_address=self.reuse_address,
                                             reuse_port=self.reuse_port, backlog=self.backlog,
                                             ssl_handshake_timeout=self.ssl_handshake_timeout)


@dataclass
class UnixSocketServer(BaseServer):
    name = "Unix Socket Server"
    peer_prefix = 'unix'
    protocol_factory: StreamServerProtocolFactory = None
    backlog: int = 100

    path: Union[str, Path] = field(default_factory=unix_address, metadata={'pickle': True})
    ssl: Optional[ServerSideSSL] = None
    ssl_handshake_timeout: Optional[int] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.ssl:
            self.close_tasks.append(self.ssl.close)
            self.ssl.set_logger(self.logger)

    @property
    def ssl_context(self) -> Optional[SSLContext]:
        if self.ssl:
            self.close_tasks.append(self.ssl.close)
            return self.ssl.context

    @property
    def listening_on(self) -> str:
        return str(self.path)

    async def _get_server(self) -> asyncio.AbstractServer:
        return await self.loop.create_unix_server(self.protocol_factory, path=str(self.path), ssl=self.ssl_context,
                                                  backlog=self.backlog, ssl_handshake_timeout=self.ssl_handshake_timeout)


class DatagramServer:

    def __init__(self, protocol_factory: DatagramServerProtocolFactory, host: str, port: int,
                 family: int = 0, proto: int = 0, flags: int = 0, sock = None, start_serving: bool = True,
                 reuse_port: bool = None, allow_broadcast: bool = None, loop = None):
        self._loop = loop
        self._sock = sock
        self._host = host
        self._port = port
        self._family = family
        self._proto = proto
        self._flags = flags
        self._reuse_port = reuse_port
        self._allow_broadcast = allow_broadcast
        self._protocol_factory = protocol_factory
        self._transport: Optional[asyncio.DatagramTransport] = None
        self._protocol: Optional[UDPServerConnection] = None
        self._status = StatusWaiter()
        if start_serving:
            asyncio.create_task(self.start_serving())
            self._status.set_starting()

    @property
    def listening_on_sockets(self) -> List[Tuple[str, int]]:
        return [self._transport.get_extra_info('sockname')]

    async def start_serving(self):
        if self.is_serving():
            return
        self._transport, self._protocol = await self._loop.create_datagram_endpoint(
            self._protocol_factory, local_addr=(self._host, self._port), family=self._family,
            allow_broadcast=self._allow_broadcast, proto=self._proto, flags=self._flags, sock=self._sock,
            reuse_port=self._reuse_port)
        self._status.set_started()

    async def wait_started(self):
        await self._status.wait_started()

    def is_serving(self) -> bool:
        return self._status.is_started()

    @property
    def sockets(self) -> List[socket.socket]:
        if self._transport:
            return [self._transport.get_extra_info('socket')]
        return []

    def get_loop(self):
        return self._loop or asyncio.get_running_loop()

    def _set_stopped(self, task: asyncio.Task):
        if task.exception():
            raise task.exception()
        self._status.set_stopped()

    def close(self):
        self._transport.close()
        self._status.set_stopped()

    async def wait_closed(self):
        await self._status.wait_stopped()


@dataclass
class UDPServer(BaseNetworkServer):
    protocol_factory: DatagramServerProtocolFactory = None
    name = "UDP Server"
    peer_prefix = 'udp'
    allow_broadcast: Optional[bool] = None

    async def _get_server(self) -> DatagramServer:
        server = DatagramServer(
            self.protocol_factory, self.host, self.port, loop=self.loop, allow_broadcast=self.allow_broadcast,
            reuse_port=self.reuse_port)
        await server.wait_started()
        return server




import io
import pickle
from dataclasses import dataclass

from aionetworking.formats.base import BaseCodec, BaseMessageObject

from typing import Any, AsyncGenerator, Tuple


@dataclass
class PickleCodec(BaseCodec):
    protocol = 4
    """
    Decode & Encode Pickle messages
    """

    async def decode(self, encoded: bytes, **kwargs) -> AsyncGenerator[Tuple[bytes, Any], None]:
        data = io.BytesIO(encoded)
        num_bytes = len(encoded)
        current_pos = 0
        while current_pos < num_bytes:
            start_pos = data.tell()
            decoded = pickle.load(data)
            current_pos = data.tell()
            yield encoded[start_pos:current_pos], decoded

    async def encode(self, decoded: Any, **kwargs) -> bytes:
        return pickle.dumps(decoded, protocol=self.protocol)


@dataclass
class PickleObject(BaseMessageObject):
    name = 'Pickle'
    codec_cls = PickleCodec



import datetime
import asyncssh
import aiofiles.os
import asyncio
from dataclasses import dataclass, field
from .exceptions import ProtocolException

from aionetworking import settings
from aionetworking.futures.schedulers import TaskScheduler
from .adaptors import ReceiverAdaptor, SenderAdaptor
from .protocol_factories import BaseProtocolFactory
from .connections import NetworkConnectionProtocol


from typing import Optional, AnyStr, Union, Dict, Any
from pathlib import Path


@dataclass
class SFTPFactory(asyncssh.SFTPServer):
    chroot = True
    remove_tmp_files = True
    base_upload_dir: Path = settings.TEMPDIR / "sftp_received"

    def __init__(self, conn, base_upload_dir: Path = settings.APP_HOME.joinpath('sftp')):
        self._scheduler = TaskScheduler()
        self._conn = conn
        self._conn.set_extra_info(sftp_factory=self)
        if self.chroot:
            root = base_upload_dir.joinpath(self._conn.get_extra_info('username'))
            root.mkdir(parents=True, exist_ok=True)
            super().__init__(self._conn, chroot=root)
        else:
            super().__init__(self._conn)

    @property
    def sftp_connection(self):
        conn = self._conn.get_extra_info('sftp_connection')
        if not conn:
            conn = self._conn.get_extra_info('connection').get_extra_info('sftp_connection')
        return conn

    async def _handle_data(self, name: str, data: AnyStr):
        self.sftp_connection.data_received(data)
        if self.remove_tmp_files:
            await aiofiles.os.remove(name)

    async def _process_completed_file(self, name: str, mode: str):
        mode = 'rb' if mode == 'wb' else 'r'
        async with aiofiles.open(name, mode=mode) as f:
            data = await f.read()
        await self._handle_data(name, data)

    def close(self, file_obj):
        super().close(file_obj)
        self._scheduler.task_with_callback(self._process_completed_file(file_obj.name, file_obj.mode))

    async def wait_closed(self):
        await self._scheduler.close()


class SFTPItem:
    def __init__(self, conn, name):
        self._conn = conn
        self._name = name
        self._value = None

    def __bool__(self):
        if not self._value:
            str(self)
        return bool(self._value)

    def __eq__(self, other):
        return str(self) == str(other)

    def __str__(self):
        if self._value:
            return self._value
        value = self._conn.get_extra_info(self._name)
        if value:
            self._value = str(value)
        return str(value)

    def __repr__(self):
        return str(self._value)


@dataclass
class BaseSFTPProtocol(NetworkConnectionProtocol):
    name = 'SFTP Server'
    conn = None
    adaptor_cls = ReceiverAdaptor
    _log_task = None

    def connection_made(self, conn: Union[asyncssh.SSHServerConnection, asyncssh.SSHClientConnection]) -> None:
        self.conn = conn
        extra_context = {
            'username': SFTPItem(self.conn, 'username')
        }
        self.logger.debug(f"""
            'client_version': {SFTPItem(self.conn, 'client_version')},
            'server_version': {SFTPItem(self.conn, 'server_version')},
            'send_cipher': {SFTPItem(self.conn, 'send_cipher')},
            'send_mac': {SFTPItem(self.conn, 'send_mac')},
            'send_compression': {SFTPItem(self.conn, 'send_compression')},
            'recv_cipher': {SFTPItem(self.conn, 'recv_cipher')},
            'recv_mac': {SFTPItem(self.conn, 'recv_mac')},
            'recv_compression': {SFTPItem(self.conn, 'recv_compression')},
        """
        )
        self.initialize_connection(conn, **extra_context)
        self.conn.set_extra_info(sftp_connection=self)

    async def wait_context_set(self) -> None:
        while not self.context.get('username'):
            await asyncio.sleep(0.000001)

    def _log_context(self, task: asyncio.Task):
        super().log_context()
        self._log_task = None

    def log_context(self):
        self._log_task = asyncio.create_task(self.wait_context_set())
        self._log_task.add_done_callback(self._log_context)

    async def _close(self, exc: Optional[BaseException]) -> None:
        if self._log_task:
            self._log_task.cancel()
        try:
            sftp_factory = self.conn.get_extra_info('sftp_factory')
            if sftp_factory:
                await sftp_factory.wait_closed()
        finally:
            await super()._close(exc)

    def close(self, immediate: bool = False):
        if not self.is_closing():
            self.conn.close()

    def connection_lost(self, exc: Optional[BaseException]) -> None:
        self.run_connection_lost_tasks()
        self.finish_connection(exc)

    def send(self, msg):
        raise ProtocolException('Unable to send messages with this receiver')


@dataclass
class SFTPServerProtocol(BaseSFTPProtocol, asyncssh.SSHServer):
    pass


@dataclass
class SFTPServerProtocolFactory(BaseProtocolFactory):
    full_name = 'SFTP Server'
    peer_prefix = 'sftp'
    connection_cls = SFTPServerProtocol


@dataclass
class SFTPClientProtocol(BaseSFTPProtocol, asyncssh.SSHClient):
    name = 'SFTP Client'
    adaptor_cls = SenderAdaptor
    sftp = None
    cwd = None
    mode = 'wb'
    strftime: str = "%Y%m%d%H%M%S%f"
    remove_tmp_files: bool = True
    prefix: str = 'FILE'
    base_path: Path = settings.TEMPDIR / "sftp_sent"
    remote_path: str = '/'
    _last_filename_timestamp: str = field(default=None, init=False)
    _last_filename_suffix: int = field(default=0, init=False)

    def __post_init__(self):
        super().__post_init__()
        self._name_lock = asyncio.Lock()
        self._last_file_name = None
        self._scheduler = TaskScheduler()
        self.base_path.mkdir(parents=True, exist_ok=True)

    async def set_sftp(self, sftp):
        self.sftp = sftp
        self.remote_path = await self.sftp.realpath(self.remote_path)

    def get_filename(self):
        timestamp = datetime.datetime.now().strftime(self.strftime)
        suffix = self._last_filename_suffix
        if timestamp == self._last_filename_timestamp:
            suffix += 1
        else:
            suffix = 0
        self._last_filename_timestamp = timestamp
        self._last_filename_suffix = suffix
        suffix = str(suffix).zfill(4)
        return self.prefix + timestamp + suffix

    async def get_tmp_path(self):
        return self.base_path / self.get_filename()

    async def _put_data(self, data: bytes):
        file_path = await self.get_tmp_path()
        self.logger.debug("Using temp path for sending file: %s", file_path)
        async with settings.FILE_OPENER(file_path, self.mode) as f:
            await f.write(data)
        await self.sftp.put(file_path, remotepath=self.remote_path)
        if self.remove_tmp_files:
            await aiofiles.os.remove(file_path)
        self.last_msg = datetime.datetime.now()

    async def wait_current_tasks(self) -> None:
        await self._adaptor.wait_current_tasks()
        await self._scheduler.wait_current_tasks()

    def send(self, data: bytes) -> asyncio.Future:
        task = self._scheduler.task_with_callback(self._put_data(data))
        return task

    async def wait_tasks_done(self) -> None:
        if self._adaptor:
            await self._adaptor.wait_current_tasks()
        await self._scheduler.close()

    async def wait_closed(self) -> None:
        await self.wait_tasks_done()
        await super().wait_closed()


@dataclass
class SFTPClientProtocolFactory(BaseProtocolFactory):
    connection_cls = SFTPClientProtocol
    remove_tmp_files: bool = True
    prefix: str = 'FILE'
    base_path: Path = settings.TEMPDIR / "sftp_sent"
    remote_path: str = '/'

    def _additional_connection_kwargs(self) -> Dict[str, Any]:
        return {
            'remove_tmp_files': self.remove_tmp_files,
            'base_path': self.base_path,
            'remote_path': self.remote_path
        }





import yaml
from .json import JSONObject
from .pickle import PickleObject
from typing import Type


def json_object_constructor(loader, node) -> Type[JSONObject]:
    return JSONObject


def load_json(Loader=yaml.SafeLoader):
    yaml.add_constructor('!JSON', json_object_constructor, Loader=Loader)


def pickle_object_constructor(loader, node) -> Type[PickleObject]:
    return PickleObject


def load_pickle(Loader=yaml.SafeLoader):
    yaml.add_constructor('!Pickle', pickle_object_constructor, Loader=Loader)


import asyncio
from dataclasses import dataclass
from aionetworking.compatibility_os import authenticate, authentication_type
from aionetworking.networking.protocol_factories import BaseProtocolFactory
from functools import partial
from .sftp import SFTPServerProtocol

from typing import Dict, Any


@dataclass
class SFTPServerOSAuthProtocol(SFTPServerProtocol):
    windows_domain: str = '.'
    pam_service: str = 'sftplogin'

    def password_auth_supported(self) -> bool:
        return True

    async def validate_password(self, username: str, password: str) -> bool:
        self.logger.info('Attempting SFTP %s login for user %s', authentication_type.lower(), username)
        authorized = await asyncio.get_running_loop().run_in_executor(None,
                                                                      partial(authenticate,
                                                                              pam_service=self.pam_service,
                                                                              domain=self.windows_domain),
                                                                      username, password)
        if authorized:
            self.logger.info('SFTP User %s successfully logged in', username)
        else:
            self.logger.error('SFTP Login failed for user %s', username)
        return authorized


@dataclass
class SFTPOSAuthProtocolFactory(BaseProtocolFactory):
    peer_prefix = 'sftp'
    connection_cls = SFTPServerOSAuthProtocol
    windows_domain: str = '.'
    pam_service: str = 'sftplogin'

    def _additional_connection_kwargs(self) -> Dict[str, Any]:
        return {
            'pam_service': self.pam_service,
            'windows_domain': self.windows_domain
        }



import signal
import asyncio
import socket
import os
import platform
import sys
from functools import partial
try:
    import psutil
except ImportError:
    psutil = None

from aionetworking.types.logging import LoggerType

from typing import Callable, Optional, Dict, Any, Coroutine
try:
    from systemd import daemon
    from systemd import journal

    def send_to_journal(*args, logger: LoggerType=None, **kwargs):
        if logger:
           logger.info(args[0])
        journal.send(*args, **kwargs)

    def send_ready():
        daemon.notify('READY=1')

    def send_status(status: str):
        daemon.notify(f'STATUS={status}')

    def send_stopping():
        daemon.notify('STOPPING=1')

    def send_reloading():
       daemon.notify('RELOADING=1')
except ImportError:
    def send_to_journal(*args, **kwargs): ...

    def send_ready():
        pass

    def send_status(status: str):
        pass

    def send_stopping():
        pass

    def send_reloading():
        pass


def loop_on_signal(signum: int, callback: Callable, logger: LoggerType = None):
    if logger:
        logger.info('Signal %s received', signum)
    callback()
    if logger:
        logger.info('Completed callback for signal %s', signum)
    if os.name != 'nt' and any(s == signum for s in (signal.SIGINT, signal.SIGTERM)):
        loop = asyncio.get_running_loop()
        loop.remove_signal_handler(signal.SIGINT)
        loop.remove_signal_handler(signal.SIGTERM)


def loop_on_user1_signal(callback: Callable, logger: LoggerType = None):
    if os.name == 'posix':
        loop = asyncio.get_running_loop()
        loop.add_signal_handler(signal.SIGUSR1, partial(loop_on_signal, signal.SIGUSR1, callback, logger))


def loop_on_close_signal(callback: Callable, logger: LoggerType = None):
    if os.name == 'posix':
        loop = asyncio.get_running_loop()
        loop.add_signal_handler(signal.SIGTERM, partial(loop_on_signal, signal.SIGTERM, callback, logger))
        loop.add_signal_handler(signal.SIGINT, partial(loop_on_signal, signal.SIGINT, callback, logger))


def loop_remove_signals():
    if os.name == 'posix':
        loop = asyncio.get_running_loop()
        loop.remove_signal_handler(signal.SIGTERM)
        loop.remove_signal_handler(signal.SIGINT)
        loop.remove_signal_handler(signal.SIGUSR1)


def send_notify_start_signal(pid: int):
    if os.name == 'posix':
        os.kill(pid, signal.SIGUSR2)


def is_wsl() -> bool:
    return os.name == 'posix' and 'microsoft' in platform.uname().release.lower()


def is_aix() -> bool:
    return sys.platform.startswith("aix")


if os.name == 'posix':
    import pamela
    authentication_type = 'PAM'

    def authenticate(username: str, password: str, pam_service: str = 'sftplogin', **kwargs) -> bool:
        try:
            pamela.authenticate(username, password, pam_service)
            return True
        except pamela.PAMError:
            return False
elif os.name == 'nt':
    import pywintypes
    import win32security
    import win32con

    authentication_type = 'WINDOWS'

    def authenticate(username: str, password: str, domain: str = '.', logon_type=win32con.LOGON32_LOGON_BATCH,
                     logon_provider=win32con.LOGON32_PROVIDER_DEFAULT, **kwargs) -> bool:
        try:
            win32security.LogonUser(username, domain, password, logon_type, logon_provider)
            return True
        except pywintypes.error:
            return False


def windows_and_is_administrator() -> bool:
    if not os.name == 'nt':
        return False
    from win32com.shell import shell
    return shell.IsUserAnAdmin()


def has_ip_address(ip: str) -> bool:
    interfaces = psutil.net_if_addrs().values()
    for i in interfaces:
        if any(a.address == ip for a in i):
            return True
    return False


def is_mac_os() -> bool:
    return sys.platform == 'darwin'


WindowsSelectorEventLoopPolicy = getattr(asyncio, 'WindowsSelectorEventLoopPolicy', None)
WindowsProactorEventLoopPolicy = getattr(asyncio, 'WindowsProactorEventLoopPolicy', None)


def is_proactor():
    policy = asyncio.get_event_loop_policy()
    return hasattr(asyncio, "ProactorEventLoop") and isinstance(policy, asyncio.WindowsProactorEventLoopPolicy)


def supports_pipe_or_unix_connections() -> bool:
    # Pipe connections are not supported for the Windows selector loop
    return hasattr(socket, 'AF_UNIX')


def is_selector() -> bool:
    policy = asyncio.get_event_loop_policy()
    if os.name == 'nt':
        return isinstance(policy, asyncio.WindowsSelectorEventLoopPolicy)
    return isinstance(policy, asyncio.DefaultEventLoopPolicy)


def is_builtin_loop() -> bool:
    return is_selector() or is_proactor()


def get_client_kwargs(happy_eyeballs_delay: Optional[float] = None,
                      interleave: Optional[int] = None) -> Dict[str, Any]:
    # Uvloop doesn't support happy_eyeballs_delay and interleave so we only add them for built-in loops
    kwargs = {}
    if is_builtin_loop():
        if happy_eyeballs_delay or interleave:
            kwargs.update({'happy_eyeballs_delay': happy_eyeballs_delay, 'interleave': interleave})
    return kwargs


async def windows_keyboard_interrupt_workaround() -> None:
    # Workaround for keyboard interrupt not working in selector event in python 3.8, still there in python 3.10
    # Not sure if still needed?
    # https://stackoverflow.com/questions/27480967/why-does-the-asyncios-event-loop-suppress-the-keyboardinterrupt-on-windows
    while True:
        await asyncio.sleep(1)


if os.name == 'nt':
    def run(coro: Coroutine, debug: bool = False):
        async def main(coro: Coroutine):
            if not is_proactor():
                asyncio.create_task(windows_keyboard_interrupt_workaround())
            await coro
        asyncio.run(main(coro), debug=debug)
else:
    run = asyncio.run



import asyncio
from collections import ChainMap
from ipaddress import AddressValueError
import builtins
import operator
import os
import re
import signal
import time
import itertools
import sys
import socket
import tempfile
from aiofiles.os import wrap
from aionetworking.compatibility_os import is_wsl, is_aix, WindowsProactorEventLoopPolicy, WindowsSelectorEventLoopPolicy
from dataclasses import dataclass, fields, MISSING
from functools import wraps

from pathlib import Path
from typing import Sequence, Callable, List, AnyStr, Tuple, Union, AsyncGenerator, Any, TYPE_CHECKING, Optional, Protocol
from ipaddress import IPv4Network, IPv6Network, IPv4Address, IPv6Address

try:
    import psutil
    supports_system_info = True
except ImportError:
    psutil = None
    supports_system_info = False


str_to_list = re.compile(r"^\s+|\s*,\s*|\s+$")


def better_file_not_found_error(*files: Union[str, Path], purpose: str = None):
    if purpose:
        additional = f'required for {purpose } '
    else:
        additional = ''
    for i, path in enumerate(files):
        path = Path(path)
        if not path.exists():
            return f'Path {path} {additional}was not found'
    return f'Unknown file {additional}was not found'


###Coroutines###
makedirs = wrap(os.makedirs)
    

def set_task_name(task: asyncio.Task, name: str, include_hierarchy: bool = True, separator: str = ':'):
    task_name = task.get_name()
    new_name = f"{task_name}_{name}" if name else task_name
    if include_hierarchy:
        prefix = get_current_task_name()
        if any(prefix == text for text in ("No Running Loop", "No Task", task_name)):
            prefix = ''
        new_name = f"{prefix}{separator}{new_name}" if prefix else new_name
    task.set_name(new_name)


def set_current_task_name(name: str, include_hierarchy: bool = True, separator: str = ':'):
    task = asyncio.current_task()
    set_task_name(task, name, include_hierarchy=include_hierarchy, separator=separator)


def get_current_task_name():
    try:
        if task := asyncio.current_task():
            return task.get_name()
        return "No Task"
    except RuntimeError:
        return 'No Running Loop'


async def time_coro(coro):
    start_time = time.time()
    await coro
    return time.time() - start_time


"""
async def benchmark(async_func: Callable, *args, num_times: int = 5, quiet: bool = False, cleanup: Callable = None,
                    cleanup_args=(), num_items: int = None, num_bytes: int = None, ignore_first: bool = True, **kwargs):
    times = []
    if not quiet:
        print("Running", async_func.__name__)
    total = 0
    if ignore_first:
        await time_coro(async_func(*args, **kwargs))
        if cleanup:
            await cleanup(*cleanup_args)
    for _ in range(0, num_times):
        time_taken = await time_coro(async_func(*args, **kwargs))
        if cleanup:
            await cleanup(*cleanup_args)
        times.append(str(time_taken))
        total += time_taken
        if not quiet:
            print(time_taken)
    average = total / num_times
    if not quiet:
        print("Average time taken:", average)
        if num_items:
            average_per_item = average / num_items
            items_per_second = 1 / average_per_item
            print('Num Items:', num_items)
            print("Average per item:", average_per_item)
            print("Items per second:", items_per_second)
            if num_bytes:
                print("Bytes per second:", (num_bytes / average))
                times = '\t'.join(times)
                print(f"{async_func.__name__}\t{num_bytes}\t{num_items}\t{times}")

"""
###Dataclasses###

def dataclass_getstate(self):
    state = {}
    f = fields(self)
    for field in f:
        name = field.name
        if field.init:
            try:
                value = getattr(self, field.name)
                if field.default == MISSING and field.default_factory == MISSING:
                    state[name] = value
                elif field.default != MISSING and value != field.default:
                    if field.metadata.get('pickle', True):
                        state[name] = value
                elif field.default_factory != MISSING:
                    if field.metadata.get('pickle', False):
                        state[name] = value
            except AttributeError:
                pass
    return state


def dataclass_setstate(self, state):
    self.__dict__.update(self.__class__(**state).__dict__)


###Typing###

class EmptyProtocol(Protocol):
    pass


def inherit_on_type_checking_only(arg):
    """Decorator for protocols/ABC classes who's methods can be used for
    type checking on subclasses but are not available at runtime.
    Used to add type hints for dynamic attributes (__getattr__, etc)
    """
    if TYPE_CHECKING:
        return arg
    return EmptyProtocol


###Async Generators###

async def aone(generator: AsyncGenerator) -> Any:
    item = await generator.__anext__()
    await generator.aclose()
    return item


async def anext(generator: AsyncGenerator) -> Any:
    return await generator.__anext__()


async def alist(generator: AsyncGenerator) -> List[Any]:
    return [i async for i in generator]


###Asyncio###

def run_in_loop(f):
    @wraps(f)
    def wrapper(*args, **kwds):
        return asyncio.run(f(*args, **kwds))

    return wrapper


def set_proactor_loop_policy_windows() -> None:
    asyncio.set_event_loop_policy(WindowsProactorEventLoopPolicy())


def set_selector_loop_policy_windows() -> None:
    asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())


def set_selector_loop_policy_posix() -> None:
    asyncio.set_event_loop_policy(asyncio.DefaultEventLoopPolicy())


def set_uvloop_policy_posix() -> None:
    import uvloop
    uvloop.install()


class InvalidLoopNameException(BaseException):
    pass


def set_loop_policy(posix_loop_type: str = None, windows_loop_type: str = None) -> None:
    if os.name == 'posix':
        if posix_loop_type == 'selector':
            set_selector_loop_policy_posix()
        elif posix_loop_type == 'uvloop':
            set_uvloop_policy_posix()
        else:
            raise InvalidLoopNameException(
                f'{posix_loop_type} is not a valid loop name. Choose either selector or uvloop')
    elif os.name == 'nt':
        if windows_loop_type == 'selector':
            set_selector_loop_policy_windows()
        elif windows_loop_type == 'proactor':
            set_proactor_loop_policy_windows()
        else:
            raise InvalidLoopNameException(
                f'{posix_loop_type} is not a valid loop name. Choose either proactor or selector')


###Pipes###

_mmap_counter = itertools.count()


def arbitrary_address(family, future_pid=False) -> Union[Tuple[str, int], Path]:
    if family == 'AF_INET':
        return 'localhost', 0
    if family == 'AF_UNIX':
        return Path(tempfile.mktemp(prefix='listener-', dir=tempfile.mkdtemp(prefix='pymp-')))
    if family == 'AF_PIPE':
        pid = '{pid}' if future_pid else os.getpid()
        return Path(tempfile.mktemp(prefix=r'\\.\pipe\pyc-%d-%d-' %
                               (pid, next(_mmap_counter)), dir=""))
    raise ValueError('unrecognized family')


def unix_address() -> Path:
    return arbitrary_address('AF_UNIX')


def pipe_address() -> Path:
    return arbitrary_address('AF_PIPE')


def pipe_address_by_os() -> Path:
    if hasattr(socket, 'AF_UNIX'):
        return unix_address()
    if sys.platform == 'win32':
        return pipe_address()
    OSError("Neither AF_UNIX nor Named Pipe is supported on this platform")


###Networking###

def hostname_or_ip(host: str):
    if os.name == 'nt':
        if host == '127.0.0.1':
            return 'localhost'
        elif host == '::1':
            return 'ip6-localhost'
    try:
        return socket.gethostbyaddr(host)[0]
    except socket.herror:
        return host


def addr_tuple_to_str(addr: Sequence):
    return ':'.join(str(a) for a in addr)


def addr_str_to_tuple(addr: AnyStr):
    addr, port = addr.split(':')
    return addr, int(port)


def ipv4_or_ipv6(address: Union[str, IPv4Address, IPv6Address]):
    try:
        return IPv4Address(address)
    except AddressValueError:
        return IPv6Address(address)


class IPNetwork:
    ip_network: Union[IPv4Network, IPv6Network] = None
    hostname: str = None
    is_ipv6: bool = False

    def __init__(self, network: Union[str, IPv4Network, IPv6Network]):
        if isinstance(network, IPv4Network):
            self.ip_network = network
        elif isinstance(network, IPv6Network):
            self.ip_network = network
            self.is_ipv6 = True
        else:
            try:
                self.ip_network = IPv4Network(network)
            except AddressValueError:
                try:
                    self.ip_network = IPv6Network(network)
                    self.is_ipv6 = True
                except AddressValueError:
                    if network:
                        self.hostname = network
                    else:
                        AddressValueError('IP Address, Network Or Hostname must be given')

    def __eq__(self, other):
        if self.ip_network:
            return self.ip_network == other.ip_network
        return self.hostname == other.hostname

    def supernet_of(self, network: Union[IPv4Network, IPv6Network], hostname: str):
        if self.ip_network:
            return self.ip_network.supernet_of(network)
        return hostname == self.hostname


def supernet_of(network: Union[str, IPNetwork, IPv4Network, IPv6Network], hostname: str, networks: Sequence[IPNetwork]):
    if not isinstance(network, IPNetwork):
        network = IPNetwork(network)
    if network.is_ipv6:
        networks = filter(lambda n: n.is_ipv6, networks)
    else:
        networks = filter(lambda n: not n.is_ipv6, networks)
    return any(n.supernet_of(network.ip_network, hostname) for n in networks)


def port_from_out(out: str) -> int:
    return int(re.search(r'(?<=:)[1-6][0-9]{3,4}', out)[0])


def wait_on_capsys(capsys) -> Tuple[str, int]:
    out = None
    while not out or "Serving" not in out:
        out = capsys.readouterr().out
        if not out or "Serving" not in out:
            time.sleep(0.1)
    port = port_from_out(out)
    return out, port


def raise_signal(signal_num: int, pid: int = None) -> None:
    num_times = 3 if os.name == 'nt' else 1
    pid = pid or os.getpid()
    for i in range(0, num_times):
        os.kill(pid, signal_num)


def wait_server_started_raise_signal(signal_num: int, host: str, capsys, pid: int = None) -> Tuple[str, int]:
    out, port = wait_on_capsys(capsys)
    assert is_listening_on((host, port), pid=pid)
    raise_signal(signal_num, pid=pid)
    return out, port


def modify_config_file(tmp_config_file, old_host, new_host):
    with open(str(tmp_config_file), "rt") as f:
        data = f.read()
        data = data.replace(old_host, new_host)
    with open(str(tmp_config_file), 'wt') as f:
        f.write(data)


def assert_reload_ok(signal_num: int, host: str, new_host: str, tmp_config_file: Path, capsys) -> Tuple[str, int]:
    out, port = wait_on_capsys(capsys)
    assert is_listening_on((host, port))
    modify_config_file(tmp_config_file, host, new_host)
    raise_signal(signal_num)
    out, new_port = wait_on_capsys(capsys)
    assert not is_listening_on((host, port))
    assert is_listening_on((new_host, new_port))
    raise_signal(signal.SIGINT)
    return out, new_port


def get_ip_port(host: str, transport) -> Tuple[str, int]:
    network = IPNetwork(host)
    if network.is_ipv6:
        actual_ip, actual_port, flowinfo, scope_id = transport.get_extra_info('sockname')
    else:
        actual_ip, actual_port = transport.get_extra_info('sockname')
    return actual_ip, actual_port


###System###


def _process_by_id(pid: int):
    return [p for p in psutil.process_iter() if p.pid == pid][0]


def is_listening_on(addr: Tuple[str, int], kind: str = 'inet', pid: int = None) -> bool:
    if psutil and not is_wsl() and not is_aix():
        pid = pid or os.getpid()
        process = _process_by_id(pid)
        connections = process.connections(kind=kind)
        return any(
            conn.laddr == addr and any(status == conn.status for status in (psutil.CONN_LISTEN, psutil.CONN_NONE)) for
            conn in connections)
    else:
        import subprocess
        process = subprocess.run(f'nc -z {addr[0]} {addr[1]}'.split())
        return process.returncode == 0


class SystemInfo:
    @property
    def memory(self):
        try:
            return psutil.Process(os.getpid()).memory_info()[0]/2.**30
        except NameError:
            return "Unknown"

    @property
    def cpu(self):
        try:
            return psutil.Process(os.getpid()).cpu_percent()
        except NameError:
            return "Unknown"


###Misc###

@dataclass
class CallableFromString(Callable):
    _strings_to_callables = {}
    callable: Callable

    def __post_init__(self):
        self.callable = self.adapt_callable(self.callable)

    def __call__(self, *args, **kwargs):
        return self.callable(*args, **kwargs)

    def adapt_callable(self, v: Union[str, Callable]):
        if isinstance(v, str):
            try:
                return self._strings_to_callables[v]
            except KeyError:
                raise TypeError(f"{v} not found")
        if v not in self._strings_to_callables.values():
            name = self.__class__.__name__
            raise TypeError(f"{v} not a valid {name}")
        return v


class Builtin(CallableFromString):
    _strings_to_callables = ChainMap(builtins.__dict__, {'istr': str})


def in_(a, b) -> bool:
    return a in b


class Operator(CallableFromString):
    _strings_to_callables = {
        '=': operator.eq,
        '==': operator.eq,
        'eq': operator.eq,
        '<': operator.lt,
        'lt': operator.lt,
        '<=': operator.le,
        'lte': operator.le,
        '!<': operator.ne,
        'ne': operator.ne,
        '>': operator.gt,
        'gt': operator.gt,
        '>=': operator.ge,
        'ge': operator.ge,
        'in': in_,
        'contains': operator.contains
    }


@dataclass
class Expression:
    attr: Optional[str]
    op: Operator
    value: Any
    adapt_expected_value: bool = True
    case_sensitive: bool = False

    @classmethod
    def from_string(cls, string: str) -> Optional['Expression']:
        """
        Generate an expression from a string, which can later be used
        for comparison purposes with the attribute of an object:
        Examples:
        expr = Expression.from_string('method = login')

        class Test:
            def __init__(self, method: str):
                self.method = method

        t1 = test('login')
        t2 = test('logout)
        expr(t1)
        True
        expr(t2)
        False

        Case insensitive:
        expr = Expression.from_string('method i= login')

        Booleans:
        expr = Expression.from_string('params')
        expr = Expression.from_string('not params')

        Value is in expected value

        expr = Expression.from_string('method in logins')
        expr(t1)
        True
        expr(t2)
        False

        Value contains expected value
        expr = Expression.from_string('method contains log')
                expr(t1)
        True
        expr(t2)
        True
        """
        case_sensitive = False
        adapt_expected_value = True
        if string:
            if ' ' not in string:
                attr = string
                op = 'eq'
                adapt_expected_value = False
                value = True
            elif string.startswith('not '):
                attr = string.split('not ')[1]
                op = 'eq'
                adapt_expected_value = False
                value = False
            else:
                attr, op, value = string.split()
                if op.startswith('i') and not op == 'in':
                    case_sensitive = True
                    value = value.lower()
                    op = op.split('i', 1)[1]
            op = Operator(op)
            return cls(attr, op, value, adapt_expected_value=adapt_expected_value, case_sensitive=case_sensitive)
        return None

    def __call__(self, obj: Any) -> bool:
        if not self.attr or self.attr == 'self':
            value = obj
        else:
            value = getattr(obj, self.attr)
        if self.case_sensitive:
            if isinstance(value, (tuple, list)):
                value = [v.lower() for v in value]
            else:
                value = value.lower()
            return self.op(value, self.value)
        if self.adapt_expected_value:
            value_type = type(value)
            compare_value = value_type(self.value)
        else:
            value_type = type(self.value)
            value = value_type(value)
            compare_value = self.value
        return self.op(value, compare_value)



from abc import abstractmethod
import asyncio

from dataclasses import dataclass, field, replace

from pathlib import Path
from typing import Tuple, Sequence, AnyStr
from aionetworking.logging.loggers import get_logger_sender
from aionetworking.types.logging import LoggerType
from aionetworking.types.networking import ProtocolFactoryType, ConnectionType
from aionetworking.utils import addr_tuple_to_str, dataclass_getstate, dataclass_setstate, run_in_loop, get_ip_port
from aionetworking.futures.value_waiters import StatusWaiter
from aionetworking.networking.connections_manager import get_unique_name
from .protocols import SenderProtocol

from typing import Callable, List, Optional, Protocol


@dataclass
class BaseSender(SenderProtocol, Protocol):
    name = 'sender'
    logger: LoggerType = field(default_factory=get_logger_sender)
    close_tasks: List[Callable] = field(default_factory=list, compare=False, hash=False, repr=False)
    _status: StatusWaiter = field(default_factory=StatusWaiter, init=False)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    def __getstate__(self):
        return dataclass_getstate(self)

    def __setstate__(self, state):
        return dataclass_setstate(self, state)

    async def __aenter__(self) -> ConnectionType:
        return await self.connect()

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        await self.close()

    async def close(self):
        if self.close_tasks:
            await asyncio.wait([asyncio.create_task(coro()) for coro in self.close_tasks])

    def is_started(self) -> bool:
        return self._status.is_started()

    def is_closing(self) -> bool:
        return self._status.is_stopping_or_stopped()

    async def wait_stopped(self) -> None:
        await self._status.wait_stopped()


@dataclass
class BaseClient(BaseSender, Protocol):
    expected_connection_exceptions = (ConnectionRefusedError,)
    name = "Client"
    peer_prefix = ''
    protocol_factory:  ProtocolFactoryType = None
    conn: ConnectionType = field(init=False, default=None)
    transport: Optional[asyncio.BaseTransport] = field(init=False, compare=False, default=None)
    timeout: int = 5

    def __post_init__(self):
        self.protocol_factory = replace(self.protocol_factory, logger=self.logger)
        self._full_name = get_unique_name(self.full_name)
        self.protocol_factory.set_name(self._full_name, self.peer_prefix)

    @abstractmethod
    async def _open_connection(self) -> ConnectionType: ...

    @property
    @abstractmethod
    def src(self) -> str: ...

    @property
    def full_name(self) -> str:
        return f"{self.name} {self.dst}"

    async def _close_connection(self) -> None:
        if self.conn and self.transport:
            await self.conn.wait_current_tasks()
            self.transport.close()
            await self.conn.wait_closed()
        self.transport = None

    def is_closing(self) -> bool:
        return self._status.is_stopping_or_stopped() or self.transport.is_closing()

    async def connect(self) -> ConnectionType:
        self._status.set_starting()
        try:
            await self.protocol_factory.start(logger=self.logger)
            self.logger.info("Opening %s connection to %s", self.name, self.dst)
            connection = await self._open_connection()
            connection.add_connection_lost_task(self.on_connection_lost)
            await connection.wait_connected()
            self._status.set_started()
            return connection
        except BaseException:
            await self.close()
            raise

    async def on_connection_lost(self) -> None:
        if not self._status.is_stopping_or_stopped():
            self.logger.info('%s connection to %s was closed on the other end', self.name, self.dst)
            self._status.set_stopping()
            await self.protocol_factory.close()
            self._status.set_stopped()

    async def close(self) -> None:
        self._status.set_stopping()
        self.logger.info("Closing %s connection to %s", self.name, self.dst)
        await self._close_connection()
        await super().close()
        await self.protocol_factory.close()
        self._status.set_stopped()

    @run_in_loop
    async def open_send_msgs(self, msgs: Sequence[AnyStr], interval: int = None, start_interval: int = 0,
                             override: dict = None, wait_responses: bool = False) -> None:
        if override:
            for k, v in override.items():
                setattr(self, k, v)
        async with self as conn:
            await asyncio.sleep(start_interval)
            for msg in msgs:
                if interval is not None:
                    await asyncio.sleep(interval)
                conn.send_data(msg)
            if wait_responses:
                for _ in msgs:
                    await conn.wait_notification()
            await asyncio.sleep(0.1) ##Workaround for bpo-38471

    @run_in_loop
    async def open_play_recording(self, path: Path, hosts: Sequence = (), timing: bool = True) -> None:
        async with self as conn:
            await conn.play_recording(path, hosts=hosts, timing=timing)


@dataclass
class BaseNetworkClient(BaseClient, Protocol):
    name = "Network Client"

    host: str = '127.0.0.1'
    port: int = 4000
    srcip: str = None
    srcport: int = 0
    _actual_src: Optional[Tuple[str, int]] = field(default=None, init=False, repr=False)

    @property
    def local_addr(self) -> Optional[Tuple[str, int]]:
        if self.srcip:
            return self.srcip, self.srcport
        return None

    @property
    def actual_local_addr(self) -> Tuple[str, int]:
        if self.host and self.transport:
            self._actual_listening_on = get_ip_port(self.host, self.transport)
            return self._actual_listening_on
        if self._actual_listening_on:
            return self._actual_listening_on

    @property
    def src(self) -> str:
        return addr_tuple_to_str(self.actual_local_addr)

    @property
    def dst(self) -> str:
        return f"{self.host}:{str(self.port)}"



import asyncio
from collections import defaultdict
from typing import DefaultDict


def future_defaultdict() -> DefaultDict:
    return defaultdict(asyncio.Future)


def queue_defaultdict() -> DefaultDict:
    return defaultdict(asyncio.Queue)




import yaml
from .sftp import SFTPClientProtocolFactory
from .sftp_os_auth import SFTPOSAuthProtocolFactory


def sftp_server_protocol_factory_constructor(loader, node) -> SFTPOSAuthProtocolFactory:
    value = loader.construct_mapping(node)
    return SFTPOSAuthProtocolFactory(**value)


def sftp_client_protocol_factory_constructor(loader, node) -> SFTPClientProtocolFactory:
    value = loader.construct_mapping(node)
    return SFTPClientProtocolFactory(**value)


def load_sftp_server_protocol_factory(Loader=yaml.SafeLoader):
    yaml.add_constructor('!SFTPServerProtocolFactory', sftp_server_protocol_factory_constructor, Loader=Loader)


def load_sftp_client_protocol_factory(Loader=yaml.SafeLoader):
    yaml.add_constructor('!SFTPClientProtocolFactory', sftp_client_protocol_factory_constructor, Loader=Loader)


import os
from pathlib import Path
import aiofiles
import tempfile
import sys


APP_NAME = 'AIONetworking'
FILE_OPENER = aiofiles.open
APP_CONFIG = {}


def __getattr__(name):
    path = None
    if name == 'TEMPDIR':
        path = Path(tempfile.gettempdir()) / sys.modules[__name__].APP_NAME.replace(" ", "")
    elif name == 'APP_HOME':
        path = Path(os.environ.get('appdata', Path.home()), sys.modules[__name__].APP_NAME)
    if path:
        path.mkdir(parents=True, exist_ok=True)
        return path



from abc import abstractmethod
import asyncio
from dataclasses import dataclass, field, replace
import os

from .exceptions import ServerException
from aionetworking.compatibility_os import loop_on_close_signal, send_ready, send_stopping, send_status, send_notify_start_signal
from aionetworking.logging.loggers import get_logger_receiver
from aionetworking.networking.connections_manager import get_unique_name
from aionetworking.types.logging import LoggerType
from aionetworking.futures.value_waiters import StatusWaiter
from aionetworking.types.networking import ProtocolFactoryType
from aionetworking.utils import dataclass_getstate, dataclass_setstate, run_in_loop, addr_tuple_to_str
import sys
from .protocols import ReceiverProtocol

from typing import Optional, Generator, List, Tuple, Callable, Protocol


@dataclass
class BaseReceiver(ReceiverProtocol, Protocol):
    name = 'receiver'
    quiet: bool = False
    logger: LoggerType = field(default_factory=get_logger_receiver)
    close_tasks: List[Callable] = field(default_factory=list, compare=False, hash=False, repr=False)
    _status: StatusWaiter = field(default_factory=StatusWaiter, init=False)

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return asyncio.get_running_loop()

    def __getstate__(self):
        return dataclass_getstate(self)

    def __setstate__(self, state):
        return dataclass_setstate(self, state)

    @run_in_loop
    async def serve_in_loop(self) -> None:
        await self.start()

    @property
    def listening_on_msgs(self) -> Generator[str, None, None]:
        yield from ()

    def send_status(self):
        msgs = ','.join(list(self.listening_on_msgs))
        if msgs:
            send_status(msgs)

    async def close(self):
        if self.close_tasks:
            await asyncio.wait([asyncio.create_task(coro()) for coro in self.close_tasks])

    async def serve_until_close_signal(self, stop_event: asyncio.Event = None,
                                       restart_event: asyncio.Event = None, notify_pid: int = None) -> None:
        stop_event = stop_event or asyncio.Event()
        restart_event = restart_event or asyncio.Event()
        loop_on_close_signal(stop_event.set, self.logger)
        await self.start()
        sys.stdout.flush()
        self.send_status()
        send_ready()
        if notify_pid:
            send_notify_start_signal(notify_pid)
        self.logger.info('Running server until signal received')
        done, pending = await asyncio.wait([asyncio.create_task(stop_event.wait()), asyncio.create_task(restart_event.wait())],
                                           return_when=asyncio.FIRST_COMPLETED)
        if stop_event.is_set():
            self.logger.info('Stop event has been set')
            send_stopping()
        if restart_event.is_set():
            self.logger.info('Restart event has been set')
        for task in pending:
            task.cancel()
        await self.close()


@dataclass
class BaseServer(BaseReceiver, Protocol):
    _serving_forever_fut = None
    name = 'Server'
    peer_prefix = 'server'

    protocol_factory:  ProtocolFactoryType = None
    server: asyncio.AbstractServer = field(default=None, init=False)

    def __post_init__(self) -> None:
        self._full_name = get_unique_name(self.full_name)
        self.protocol_factory = replace(self.protocol_factory, quiet=self.quiet)
        self.protocol_factory.set_name(self._full_name, self.peer_prefix)

    @property
    @abstractmethod
    def listening_on(self) -> str: ...

    @property
    def full_name(self):
        return f"{self.name} {self.listening_on}"

    @property
    def listening_on_sockets(self) -> List[Tuple[str, int]]:
        return [socket.getsockname() for socket in self.server.sockets]

    @property
    def listening_on_msgs(self) -> Generator[str, None, None]:
        for sock_name in self.listening_on_sockets:
            if isinstance(sock_name, (tuple, list)):
                listening_on = ':'.join([str(v) for v in sock_name[0:2]])
            else:
                listening_on = sock_name
            yield f"Serving {self.name} on {listening_on}"

    def _print_listening_message(self) -> None:
        for msg in self.listening_on_msgs:
            print(msg)

    async def start(self) -> None:
        if self._status.is_starting_or_started():
            raise ServerException(f"{self.name} running on {self.listening_on} already started")
        self._status.set_starting()
        try:
            await self.protocol_factory.start(logger=self.logger)
            self.logger.info('Starting %s on %s', self.name, self.listening_on)
            await self._start_server()
            if not self.quiet:
                self._print_listening_message()
            self._status.set_started()
        except Exception:
            await self.close()
            raise

    async def _serve_forever(self) -> None:
        if self._serving_forever_fut is not None:
            raise RuntimeError(
                f'server {self!r} is already being awaited on serve_forever()')
        if not self._status.is_starting_or_started():
            await self.start()
        self._serving_forever_fut = self.loop.create_future()

    async def serve_forever(self) -> None:
        if not self._status.is_starting_or_started():
            await self.start()
        try:
            if hasattr(self.server, 'serve_forever'):
                self._serving_forever_fut = asyncio.ensure_future(self.server.serve_forever())

            else:
                await self._serve_forever()
            try:
                await self._serving_forever_fut
            finally:
                self._serving_forever_fut = None
        except asyncio.CancelledError:
            try:
                await self.close()
            finally:
                raise

    async def close(self) -> None:
        if self._status.is_stopping_or_stopped():
            return None
        self._status.set_stopping()
        self.logger.info('Stopping %s running at %s', self.name, self.listening_on)
        if (self._serving_forever_fut is not None and
                not self._serving_forever_fut.done()):
            self._serving_forever_fut.cancel()
            self._serving_forever_fut = None
        await self._stop_server()
        await super().close()
        self.logger.info('Stopping protocol factory')
        await self.protocol_factory.close()
        self.logger.info('%s stopped', self.name)
        self._status.set_stopped()

    def close_all_connections(self) -> None:
        self.protocol_factory.close_all_connections(None)

    async def wait_num_connections(self, num: int):
        await self.protocol_factory.wait_num_connected(num)

    async def wait_num_has_connected(self, num: int):
        await self.protocol_factory.wait_num_has_connected(num)

    async def wait_all_connections_closed(self):
        await self.protocol_factory.wait_all_closed()

    async def wait_all_tasks_done(self) -> None:
        await self.protocol_factory.close_actions()

    def is_started(self) -> bool:
        return self._status.is_started()

    def is_closing(self) -> bool:
        return self._status.is_stopping_or_stopped()

    def is_closed(self) -> bool:
        return self._status.is_stopped()

    async def wait_started(self):
        await self._status.wait_started()

    async def wait_has_started(self):
        await self._status.wait_has_started()

    async def wait_stopped(self):
        await self._status.wait_stopped()

    @abstractmethod
    async def _get_server(self) -> asyncio.AbstractServer: ...

    async def _start_server(self) -> None:
        self.server = await self._get_server()

    async def _stop_server(self) -> None:
        if self.server:
            self.server.close()
            await self.server.wait_closed()


@dataclass
class BaseNetworkServer(BaseServer, Protocol):
    host: str = '0.0.0.0'
    port: int = 4000
    reuse_port: Optional[bool] = None
    backlog: int = 100
    _actual_listening_on: Optional[Tuple[str, int]] = field(default=None, init=False, repr=False)

    @property
    def listening_on(self) -> str:
        local_addr = self.actual_local_addr
        local_addr = local_addr or (self.host, self.port)
        return addr_tuple_to_str(local_addr)

    @property
    def actual_local_addr(self) -> Optional[Tuple[str, int]]:
        if self.is_started():
            self._actual_listening_on = self.listening_on_sockets[0]
            return self._actual_listening_on
        if self._actual_listening_on:
            return self._actual_listening_on



import asyncio
from dataclasses import dataclass, field
from typing import Any, DefaultDict, Dict

from aionetworking.factories import future_defaultdict


@dataclass
class Counter:
    _num: int = 0
    _total_increments: int = 0
    _num_waiters: DefaultDict[int, asyncio.Future] = field(default_factory=future_defaultdict, init=False)
    _total_increment_waiters: DefaultDict[int, asyncio.Future] = field(default_factory=future_defaultdict, init=False)
    max: int = None
    max_increments: int = None
    aliases: Dict[str, int] = field(default_factory=dict)

    def __next__(self):
        self.increment()
        return self._num

    def increment(self) -> int:
        if self.max and self._num == self.max:
            raise ValueError(f'counter incrementing above the maximum value: {self.max}')
        if self.max_increments and self.max_increments == self._total_increments:
            raise ValueError(f'counter incremented more times than allowed: {self._total_increments}')
        self._num += 1
        self._total_increments += 1
        self._check_num_waiters()
        self._check_total_increment_waiters()
        return self._num

    def decrement(self) -> int:
        if self._num <= 0:
            raise ValueError('counter decremented too many times')
        self._num -= 1
        self._check_num_waiters()
        return self._num

    @property
    def num(self):
        return self._num

    @property
    def total_increments(self):
        return self._total_increments

    def _check_num_waiters(self):
        if self._num in self._num_waiters:
            self._num_waiters[self._num].set_result(True)
            fut = self._num_waiters.pop(self._num)
            assert fut.done()

    def _check_total_increment_waiters(self):
        if self._total_increments in self._total_increment_waiters:
            self._total_increment_waiters[self._total_increments].set_result(True)
            fut = self._total_increment_waiters.pop(self._total_increments)
            assert fut.done()

    async def _wait_for(self, num: int, target_num: int, waiters: DefaultDict[int, asyncio.Future]):
        if num == target_num:
            return
        fut = waiters[num]
        await fut

    async def wait_for(self, num: int) -> None:
        if num == self._num:
            return
        return await self._wait_for(num, self._num, self._num_waiters)

    def set_num(self, num: int):
        self._num = num
        self._check_num_waiters()
        return self._num

    async def wait(self, alias: str):
        num = self.aliases[alias]
        await self.wait_for(num)

    def set(self, alias: str):
        num = self.aliases[alias]
        self.set_num(num)

    async def wait_for_total_increments(self, num: int) -> None:
        if num == self._total_increments:
            return
        return await self._wait_for(num, self._total_increments, self._total_increment_waiters)


class Counters(DefaultDict[Any, Counter]):
    default_factory = Counter

    def __missing__(self, key):
        result = self[key] = self.default_factory()
        return result

    def increment(self, key: Any) -> int:
        return self[key].increment()

    def decrement(self, key: Any) -> int:
        return self[key].decrement()

    async def wait_for(self, key: Any, num: int):
        counter = self[key]
        await counter.wait_for(num)

    async def wait_for_total_increments(self, key: Any, num: int) -> None:
        counter = self[key]
        await counter.wait_for_total_increments(num)

    def total_increments(self, key: Any) -> int:
        return self[key].total_increments

    def get_num(self, key: Any) -> int:
        return self[key].num

    def remove(self, key: Any) -> None:
        self.pop(key, None)



class MessageFromNotAuthorizedHost(BaseException):
    pass


class MethodNotFoundError(BaseException):
    pass


class ConnectionAlreadyClosedError(BaseException):
    pass


class ProtocolException(BaseException):
    pass


class RemoteConnectionClosedError(BaseException):
    pass


from dataclasses import dataclass, field

from aionetworking.logging.loggers import get_logger_receiver
from aionetworking.types.logging import LoggerType
from aionetworking.types.formats import MessageObjectType
from aionetworking.futures.value_waiters import StatusWaiter
from aionetworking.utils import dataclass_getstate, dataclass_setstate
from .protocols import ActionProtocol

from typing import Any, TypeVar, AsyncGenerator, Protocol


ActionType = TypeVar('ActionType', bound='BaseAction')


@dataclass
class BaseAction(ActionProtocol, Protocol):
    supports_notifications = False
    name = 'receiver action'
    logger: LoggerType = field(default_factory=get_logger_receiver, metadata={'pickle': True})
    task_timeout: int = 10
    _status: StatusWaiter = field(default_factory=StatusWaiter, compare=False, repr=False)

    timeout: int = 5

    def _set_logger(self, logger: LoggerType = None) -> None:
        parent_logger = logger or self.logger
        self.logger = parent_logger.get_child(name='actions')

    async def start(self, logger: LoggerType = None) -> None:
        self._set_logger(logger=logger)

    def __getstate__(self):
        return dataclass_getstate(self)

    def __setstate__(self, state):
        dataclass_setstate(self, state)

    def filter(self, msg: MessageObjectType) -> bool:
        return msg.filter()

    async def get_notifications(self, peer: str) -> AsyncGenerator[None, None]:
        yield

    def is_closing(self) -> None:
        return self._status.is_stopping_or_stopped()

    async def close(self) -> None:
        self._status.set_stopped()

    def on_decode_error(self, data: bytes, exc: BaseException) -> Any:
        pass

    def on_exception(self, msg: MessageObjectType, exc: BaseException) -> Any:
        pass

    async def do_one(self, msg: MessageObjectType) -> Any: ...


@dataclass
class EmptyAction(BaseAction): ...


import asyncio
from aionetworking.actions.base import BaseAction
from aionetworking.types.formats import MessageObjectType
from dataclasses import dataclass, field
from aionetworking.factories import queue_defaultdict

from typing import AsyncGenerator, DefaultDict


class InvalidRequestError(Exception):
    pass


@dataclass
class EchoAction(BaseAction):
    supports_notifications = True
    _queues: DefaultDict[str, asyncio.Queue] = field(default_factory=queue_defaultdict, init=False, compare=False, repr=False)

    async def get_notifications(self, peer: str) -> AsyncGenerator[None, None]:
        queue = self._queues[peer]
        while True:
            item = await queue.get()
            yield {'result': item}

    def on_decode_error(self, data: bytes, exc: BaseException) -> dict:
        return {'error': 'JSON was invalid'}

    def on_exception(self, msg: MessageObjectType, exc: BaseException) -> dict:
        return {'id': msg.decoded['id'], 'error': exc.__class__.__name__}

    async def do_one(self, msg: MessageObjectType) -> dict:
        method = msg.decoded['method']
        if method == 'echo':
            id_ = msg.decoded.get('id')
            return {'id': id_, 'result': 'echo'}
        elif method == 'send_notification':
            self._queues[msg.full_sender].put_nowait('notification')
        elif method == 'simple':
            pass
        else:
            raise InvalidRequestError(f'{method} is not recognised as a valid method')



from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from aionetworking.requesters.protocols import RequesterProtocol


RequesterType = TypeVar('RequesterType', bound='RequesterProtocol')





from typing import TYPE_CHECKING, TypeVar

if TYPE_CHECKING:
    from aionetworking.formats.protocols import MessageObject, Codec
    from aionetworking.formats.contrib.json import JSONObject


MessageObjectType = TypeVar('MessageObjectType', bound='MessageObject')
CodecType = TypeVar('CodecType', bound='Codec')
JSONObjectType = TypeVar('JSONObjectType', bound='JSONObject')


from ssl import (SSLContext, Purpose, CERT_REQUIRED, CERT_NONE, PROTOCOL_TLS_CLIENT, PROTOCOL_TLS_SERVER,
    get_default_verify_paths, cert_time_to_seconds)
import asyncio
import datetime
import os
import sys
from aionetworking.logging.loggers import get_logger_receiver
from aionetworking.types.logging import LoggerType
from aionetworking.utils import better_file_not_found_error, set_task_name

from functools import cached_property
from pathlib import Path
from typing import Optional, Dict, Any, Tuple, Protocol
from dataclasses import dataclass, field


try:
    import cryptography
    warn_if_expires_before_days_default = 7
except ImportError:
    warn_if_expires_before_days_default = None


def ssl_cert_time_to_datetime(timestamp: str) -> datetime.datetime:
    return datetime.datetime.utcfromtimestamp(cert_time_to_seconds(timestamp))


def check_ssl_cert_expired(expiry_time: datetime.datetime, warn_before_days: int) -> Optional[int]:
    expires_in = (expiry_time - datetime.datetime.now()).days
    if expires_in < warn_before_days:
        return expires_in
    return None


def check_peercert_expired(peercert: Dict[str, Any], warn_before_days: int) -> Tuple[datetime.datetime, Optional[int]]:
    expiry_time = ssl_cert_time_to_datetime(peercert['notAfter'])
    cert_expiry_in_days = check_ssl_cert_expired(expiry_time, warn_before_days)
    return expiry_time, cert_expiry_in_days


@dataclass
class BaseSSLContext(Protocol):
    purpose: str = field(init=False, default=None)
    protocol: int = field(init=False, default=None)
    logger: LoggerType = field(default_factory=get_logger_receiver)
    ssl: bool = True
    cert: Path = None
    key: Path = None
    key_password: str = None
    cafile: Path = None
    capath: Path = None
    cadata: str = None
    cert_required: bool = False
    check_hostname: bool = True
    warn_if_expires_before_days: int = warn_if_expires_before_days_default
    _warn_expiry_task: asyncio.Task = field(default=None, init=False, compare=False)

    def set_logger(self, logger: LoggerType) -> None:
        self.logger = logger

    async def close(self) -> None:
        if self._warn_expiry_task and not self._warn_expiry_task.done():
            self._warn_expiry_task.cancel()

    async def check_cert_expiry(self):
        try:
            from .ssl_utils import load_cert_expiry_time
            cert_expiry_time = load_cert_expiry_time(self.cert)
            if cert_expiry_time:
                while True:
                    cert_expiry_days = check_ssl_cert_expired(cert_expiry_time, self.warn_if_expires_before_days)
                    if cert_expiry_days:
                        self.logger.warn_on_cert_expiry('Own', cert_expiry_days, cert_expiry_time)
                    await asyncio.sleep(86400)
        except ImportError:
            self.logger.warning(
                'Unable to check ssl cert validity. Install cryptography library to enable this or set warn_if_expires_before_days to 0')

    @cached_property
    def context(self) -> Optional[SSLContext]:
        if self.ssl:
            self.logger.info("Setting up SSL")
            context = SSLContext(self.protocol)
            if self.cert and self.key:
                self.logger.info("Using SSL Cert: %s", self.cert)
                try:
                    context.load_cert_chain(str(self.cert), str(self.key), password=self.key_password)
                except FileNotFoundError as e:
                    raise FileNotFoundError(better_file_not_found_error(self.cert, self.key, purpose='ssl cert loading'))
                if self.warn_if_expires_before_days:
                    self._warn_expiry_task = asyncio.create_task(self.check_cert_expiry())
                    set_task_name(self._warn_expiry_task, 'CheckSSLCertValidity')
            context.verify_mode = CERT_REQUIRED if self.cert_required else CERT_NONE
            context.check_hostname = self.check_hostname
            self.logger.info('%s, Check Hostname: %s' % (context.verify_mode, context.check_hostname))

            if context.verify_mode != CERT_NONE:
                 if self.cafile or self.capath or self.cadata:
                    locations = {
                        'cafile': str(self.cafile) if self.cafile else None,
                        'capath': str(self.capath) if self.capath else None,
                        'cadata': self.cadata
                    }
                    try:
                        context.load_verify_locations(**locations)
                        self.logger.info("Verifying SSL certs with: %s", locations)
                    except FileNotFoundError:
                        raise FileNotFoundError(
                            better_file_not_found_error(*locations.values(), purpose='CA ssl cert validation'))
                 else:
                    context.load_default_certs(self.purpose)
                    self.logger.info("Verifying SSL certs with: %s", get_default_verify_paths())
            self.logger.info("SSL Context loaded")
            # OpenSSL 1.1.1 keylog file
            if hasattr(context, 'keylog_filename'):
                keylogfile = os.environ.get('SSLKEYLOGFILE')
                if keylogfile and not sys.flags.ignore_environment:
                    self.logger.warning("TLS encryption secrets are being stored in %s", keylogfile)
                    context.keylog_filename = keylogfile
            return context
        return None


@dataclass
class ServerSideSSL(BaseSSLContext):
    protocol = PROTOCOL_TLS_SERVER
    purpose = Purpose.CLIENT_AUTH
    check_hostname: bool = False
    cert_required: bool = False


@dataclass
class ClientSideSSL(BaseSSLContext):
    protocol = PROTOCOL_TLS_CLIENT
    purpose = Purpose.SERVER_AUTH
    check_hostname: bool = True
    cert_required: bool = True





import yaml
from .sftp import SFTPClient


def sftp_client_constructor(loader, node) -> SFTPClient:
    value = loader.construct_mapping(node) if node.value else {}
    return SFTPClient(**value)


def load_sftp_client(Loader=yaml.SafeLoader):
    yaml.add_constructor('!SFTPClient', sftp_client_constructor, Loader=Loader)



import os
import re
import yaml
from pathlib import Path

from functools import partial
from aionetworking.compatibility import default_server_port, default_client_port
from aionetworking.utils import IPNetwork
from aionetworking.logging import Logger

from typing import Optional, Dict, Union, Sequence


def port_constructor(default_port_func, loader, node) -> int:
    if node.value:
        value = loader.construct_yaml_int(node)
        if value:
            return value
    return default_port_func()


def load_default_ports():
    yaml.add_constructor('!Port', partial(port_constructor, default_server_port), Loader=yaml.SafeLoader)
    yaml.add_constructor('!ClientPort', partial(port_constructor, default_client_port), Loader=yaml.SafeLoader)


def path_constructor(paths, loader, node) -> Optional[Path]:
    value = loader.construct_scalar(node)
    if value:
        path = Path.resolve(Path(value.format(**paths)))
        return path
    return None


def load_path(paths: Dict[str, Union[str, Path]], Loader=yaml.SafeLoader):
    yaml.add_constructor('!Path', partial(path_constructor, paths), Loader=Loader)


def base_path_constructor(base_path: Path, loader, node) -> Path:
    value = loader.construct_scalar(node)
    return Path(base_path / value)


def load_base_path(tag_name: str, base_path: Path, Loader=yaml.SafeLoader):
    yaml.add_constructor(tag_name, partial(base_path_constructor, base_path), Loader=Loader)


def ip_network_constructor(loader, node) -> Sequence[IPNetwork]:
    values = loader.construct_sequence(node)
    return [IPNetwork(v) for v in values]


def load_ip_network(Loader=yaml.SafeLoader):
    yaml.add_constructor('!IPNetwork', ip_network_constructor, Loader=Loader)


def logger_constructor(loader, node) -> Logger:
    value = loader.construct_mapping(node) if node.value else {}
    return Logger(**value)


def load_logger(Loader=yaml.SafeLoader):
    yaml.add_constructor('!Logger', logger_constructor, Loader=Loader)


def receiver_logger_constructor(loader, node) -> Logger:
    value = loader.construct_mapping(node) if node.value else {}
    return Logger('receiver', **value)


def load_receiver_logger(Loader=yaml.SafeLoader):
    yaml.add_constructor('!ReceiverLogger', receiver_logger_constructor, Loader=Loader)


def sender_logger_constructor(loader, node) -> Logger:
    value = loader.construct_mapping(node) if node.value else {}
    return Logger('sender', **value)


def load_sender_logger(Loader=yaml.SafeLoader):
    yaml.add_constructor('!SenderLogger', sender_logger_constructor, Loader=Loader)


env_variable_pattern = re.compile(r'.*?\${(\w+)}.*?')


def constructor_env_variables(loader, node):
    """
    Extracts the environment variable from the node's value
    :param yaml.Loader loader: the yaml loader
    :param node: the current node in the yaml
    :return: the parsed string that contains the value of the environment
    variable
    """
    value = loader.construct_scalar(node)
    match = env_variable_pattern.findall(value)
    if match:
        full_value = value
        for g in match:
            full_value = full_value.replace(
                f'${{{g}}}', os.environ.get(g, '')
            )
        return full_value
    return value


def load_env_variable(Loader=yaml.SafeLoader):
    Loader.add_implicit_resolver("!ENV", env_variable_pattern, None)
    yaml.add_constructor('!ENV', constructor_env_variables, Loader=Loader)




import yaml
from .echo import EchoRequester


def echo_requester_constructor(loader, node) -> EchoRequester:
    value = loader.construct_mapping(node) if node.value else {}
    return EchoRequester(**value)


def load_echo_requester(Loader=yaml.SafeLoader):
    yaml.add_constructor('!EchoRequester', echo_requester_constructor, Loader=Loader)




import yaml
from .clients import TCPClient, UDPClient, UnixSocketClient


def tcp_client_constructor(loader, node) -> TCPClient:
    value = loader.construct_mapping(node) if node.value else {}
    return TCPClient(**value)


def load_tcp_client(Loader=yaml.SafeLoader):
    yaml.add_constructor('!TCPClient', tcp_client_constructor, Loader=Loader)


def udp_client_constructor(loader, node) -> UDPClient:
    value = loader.construct_mapping(node) if node.value else {}
    return UDPClient(**value)


def load_udp_client(Loader=yaml.SafeLoader):
    yaml.add_constructor('!UDPClient', udp_client_constructor, Loader=Loader)


def pipe_client_constructor(loader, node):
    value = loader.construct_mapping(node) if node.value else {}
    return UnixSocketClient(**value)


def load_pipe_client(Loader=yaml.SafeLoader):
    yaml.add_constructor('!PipeClient', pipe_client_constructor, Loader=Loader)




from .receivers import TCPServer, UDPServer, UnixSocketServer
from .senders import TCPClient, UDPClient, UnixSocketClient
from .networking import (StreamServerProtocolFactory, StreamClientProtocolFactory, DatagramServerProtocolFactory,
                         DatagramClientProtocolFactory, ServerSideSSL, ClientSideSSL)
from .actions import FileStorage, BufferedFileStorage
from .logging import Logger
from .futures import TaskScheduler, Counters, Counter, ValueWaiter
from .formats import JSONObject, JSONCodec




from abc import abstractmethod
import asyncio
from dataclasses import dataclass, field
import datetime
from functools import partial

from .exceptions import MethodNotFoundError, RemoteConnectionClosedError
from aionetworking.actions.protocols import ActionProtocol
from aionetworking.utils import set_task_name
from aionetworking.logging.loggers import ConnectionLogger, get_connection_logger_receiver
from aionetworking.logging.utils_logging import p
from aionetworking.types.formats import MessageObjectType, CodecType
from aionetworking.types.networking import BaseContext
from aionetworking.formats.recording import BufferObject, BufferCodec, get_recording_from_file
from aionetworking.requesters.protocols import RequesterProtocol
from aionetworking.futures.schedulers import TaskScheduler

from .protocols import AdaptorProtocol

from pathlib import Path
from typing import Any, Callable, Generator, Dict, Sequence, Type, AsyncIterator, Optional, Protocol


def not_implemented_callable(*args, **kwargs) -> None:
    raise NotImplementedError


@dataclass
class BaseAdaptorProtocol(AdaptorProtocol, Protocol):
    dataformat: Type[MessageObjectType]
    bufferformat: Type[BufferObject] = BufferObject
    codec: CodecType = None
    buffer_codec: BufferCodec = None
    _scheduler: TaskScheduler = field(default_factory=TaskScheduler, init=False, hash=False, compare=False, repr=False)
    logger: ConnectionLogger = field(default_factory=get_connection_logger_receiver, compare=False, hash=False)
    context: BaseContext = field(default_factory=dict)
    codec_config: Dict[str, Any] = field(default_factory=dict, metadata={'pickle': True})
    preaction: ActionProtocol = None
    send: Callable[[bytes], Optional[asyncio.Future]] = field(default=not_implemented_callable, repr=False, compare=False)

    def __post_init__(self) -> None:
        self.logger.new_connection()

    def on_msg_sent(self, msg_encoded: bytes, task: Optional[asyncio.Future]):
        self.logger.on_msg_sent(msg_encoded)

    def send_data(self, msg_encoded: bytes) -> Optional[asyncio.Future]:
        self.logger.on_sending_encoded_msg(msg_encoded)
        fut = self.send(msg_encoded)
        if fut:
            fut.add_done_callback(partial(self.on_msg_sent, msg_encoded))
            return fut
        else:
            self.on_msg_sent(msg_encoded, None)

    def _set_codecs(self, buffer: Optional[bytes]):
        self.codec = self.dataformat.get_codec(buffer, logger=self.logger, context=self.context, **self.codec_config)
        self.buffer_codec: BufferCodec = self.bufferformat.get_codec(buffer, context=self.context, logger=self.logger)

    def on_encode_task_finished(self, task: asyncio.Future):
        exception = task.exception()
        if exception:
            self.logger.manage_error(exception)
            fut = None
        else:
            msg_obj = task.result()
            self.logger.on_sending_decoded_msg(msg_obj)
            fut = self.send_data(msg_obj.encoded)
        if fut:
            fut.add_done_callback(self._scheduler.task_done)
        else:
            self._scheduler.task_done(task)

    def encode_and_send_msg(self, decoded: Any) -> None:
        if not self.codec:
            self._set_codecs(decoded)

        self._scheduler.task_with_callback(self.codec.encode_obj(decoded), callback=self.on_encode_task_finished)

    def encode_and_send_msgs(self, decoded_msgs: Sequence[Any]) -> None:
        for decoded_msg in decoded_msgs:
            self.encode_and_send_msg(decoded_msg)

    async def _run_preaction(self, buffer: bytes, timestamp: datetime.datetime = None) -> None:
        self.logger.info('Running preaction')
        buffer_obj = await self.buffer_codec.encode_obj(buffer, system_timestamp=timestamp)
        if not self.preaction.filter(buffer_obj):
            await self.preaction.do_one(buffer_obj)

    def on_data_received(self, buffer: bytes, timestamp: datetime.datetime = None) -> asyncio.Future:
        timestamp = timestamp or datetime.datetime.now()
        if not self.codec:
            self._set_codecs(buffer)
        self.logger.on_buffer_received(buffer)
        if self.preaction:
            self._scheduler.task_with_callback(self._run_preaction(buffer, timestamp),
                                               name=f"{self.context['peer']}-Preaction")
        msgs_generator = self.codec.decode_buffer(buffer, system_timestamp=timestamp)
        task = self._scheduler.task_with_callback(self.process_msgs(msgs_generator, buffer), name='Process_Msgs')
        return task

    async def wait_current_tasks(self) -> None:
        await self._scheduler.wait_current_tasks()

    async def close(self, exc: Optional[BaseException] = None) -> None:
        task_count = self._scheduler.task_count
        if task_count:
            self.logger.info('Connection waiting on %s to complete', p.no('task', task_count))
        await self._scheduler.close()
        self.logger.connection_finished(exc)

    @abstractmethod
    async def process_msgs(self, msgs: AsyncIterator[MessageObjectType], buffer: bytes) -> None: ...


@dataclass
class SenderAdaptor(BaseAdaptorProtocol):
    is_receiver = False
    requester: RequesterProtocol = None

    _notification_queue: asyncio.Queue = field(default_factory=asyncio.Queue, init=False, repr=False, hash=False,
                                               compare=False)

    def __getattr__(self, item):
        if item.startswith('_'):            # Required for pytest which looks for __attrs_attrs__
            return
        if item in getattr(self.requester, 'methods', ()):
            return partial(self._run_method_and_wait, getattr(self.requester, item))
        if item in getattr(self.requester, 'notification_methods', ()):
            return partial(self._run_method, getattr(self.requester, item))
        raise MethodNotFoundError(f'{item} method was not found')

    async def wait_notification(self) -> MessageObjectType:
        return await self._notification_queue.get()

    def get_notification(self) -> MessageObjectType:
        return self._notification_queue.get_nowait()

    def all_notifications(self) -> Generator[MessageObjectType, None, None]:
        for i in range(0, self._notification_queue.qsize()):
            yield self._notification_queue.get_nowait()

    async def close(self, exc: Optional[BaseException] = None) -> None:
        self._scheduler.cancel_all_futures(RemoteConnectionClosedError)
        await super().close(exc)

    async def send_data_and_wait(self, request_id: Any, encoded: bytes) -> Any:
        return await self._scheduler.run_wait_fut(request_id, self.send_data, encoded)

    async def send_msg_and_wait(self, msg_obj: MessageObjectType) -> asyncio.Future:
        return await self.send_data_and_wait(msg_obj.request_id, msg_obj.encoded)

    async def encode_send_wait(self, decoded: Any) -> asyncio.Future:
        if not self.codec:
            self._set_codecs(decoded)
        msg_obj = await self.codec.encode_obj(decoded)
        self.logger.on_sending_decoded_msg(msg_obj)
        return await self.send_msg_and_wait(msg_obj)

    def _run_method(self, method: Callable, *args, **kwargs) -> None:
        decoded = method(*args, **kwargs)
        self.encode_and_send_msg(decoded)

    async def _run_method_and_wait(self, method: Callable, *args, **kwargs) -> asyncio.Future:
        decoded = method(*args, **kwargs)
        return await self.encode_send_wait(decoded)

    async def play_recording(self, file_path: Path, hosts: Sequence = (), timing: bool = True) -> None:
        prev_timestamp = None
        self.logger.debug("Playing recording from file %s", file_path)
        futs = []
        async for packet in get_recording_from_file(file_path):
            if (not hosts or packet.sender in hosts) and not packet.sent_by_server:
                if timing:
                    if prev_timestamp:
                        timedelta = packet.timestamp - prev_timestamp
                        seconds = timedelta.total_seconds()
                        await asyncio.sleep(seconds)
                    prev_timestamp = packet.timestamp
                fut = self.send_data(packet.data)
                if fut:
                    futs.append(fut)
        if futs:
            await asyncio.gather(*futs)
        self.logger.debug("Recording finished")

    async def process_msgs(self, msgs: AsyncIterator[MessageObjectType], buffer: bytes) -> None:
        async for msg in msgs:
            if msg.request_id is not None:
                try:
                    self._scheduler.set_result(msg.request_id, msg)
                except KeyError:
                    self._notification_queue.put_nowait(msg)
            else:
                self._notification_queue.put_nowait(msg)


@dataclass
class ReceiverAdaptor(BaseAdaptorProtocol):
    is_receiver = True
    action: ActionProtocol = None

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.action.supports_notifications:
            self._notifications_task = self._scheduler.task_with_callback(self._send_action_notifications(),
                                                                          continuous=True, name='Notifications')
        else:
            self._notifications_task = None

    def _set_codecs(self, buffer: Optional[bytes]):
        super()._set_codecs(buffer)
        if self.codec.supports_notifications and not self._notifications_task:
            self._notifications_task = self._scheduler.task_with_callback(self._send_codec_notifications(), continuous=True,
                                                                          name='Notifications')

    async def close(self, exc: Optional[BaseException] = None) -> None:
        if self._notifications_task:
            self._notifications_task.cancel()
        await super().close(exc)

    async def _send_action_notifications(self):
        async for item in self.action.get_notifications(self.context['peer']):
            self.encode_and_send_msg(item)

    async def _send_codec_notifications(self):
        async for item in self.codec.get_notifications():
            self.encode_and_send_msg(item)

    def _on_exception(self, exc: BaseException, msg_obj: MessageObjectType) -> None:
        self.logger.on_msg_failed(msg_obj, exc)
        response = self.action.on_exception(msg_obj, exc)
        if response:
            self.encode_and_send_msg(response)

    def _on_success(self, result: Any, msg_obj: MessageObjectType) -> None:
        try:
            if result:
                self.encode_and_send_msg(result)
        finally:
            self.logger.on_msg_processed(msg_obj)

    def _on_decoding_error(self, buffer: bytes, exc: BaseException):
        self.logger.manage_decode_error(buffer, exc)
        response = self.action.on_decode_error(buffer, exc)
        if response:
            self.encode_and_send_msg(response)

    async def _process_msg(self, msg_obj):
        self.logger.debug('Processing message %s', msg_obj)
        try:
            result = await self.action.do_one(msg_obj)
            self._on_success(result, msg_obj)
        except BaseException as e:
            self._on_exception(e, msg_obj)
            raise

    async def process_msgs(self, msgs: AsyncIterator[MessageObjectType], buffer: bytes) -> None:
        tasks = []
        try:
            async for msg_obj in msgs:
                if not self.action.filter(msg_obj):
                    task = asyncio.create_task(self._process_msg(msg_obj))
                    set_task_name(task, f'Process {msg_obj}')
                    tasks.append(task)
                else:
                    self.logger.on_msg_filtered(msg_obj)
            await asyncio.wait_for(asyncio.gather(*tasks), timeout=self.action.task_timeout)
        except Exception as exc:
            self._on_decoding_error(buffer, exc)
            raise




from typing import TYPE_CHECKING, TypeVar, TypedDict

if TYPE_CHECKING:
    from aionetworking.networking.protocols import (ProtocolFactoryProtocol, ConnectionProtocol,
                                                    NetworkConnectionProtocol, UDPConnectionProtocol, AdaptorProtocol,
                                                    SenderAdaptorProtocol, SimpleNetworkConnectionProtocol)


ProtocolFactoryType = TypeVar('ProtocolFactoryType', bound='ProtocolFactoryProtocol')
ConnectionType = TypeVar('ConnectionType', bound='ConnectionProtocol')
NetworkConnectionType = TypeVar('NetworkConnectionType', bound='NetworkConnectionProtocol')
UDPConnectionType = TypeVar('UDPConnectionType', bound='UDPConnectionProtocol')
AdaptorType = TypeVar('AdaptorType', bound='AdaptorProtocol')
SenderAdaptorType = TypeVar('SenderAdaptorType', bound='SenderAdaptorProtocol')
SimpleNetworkConnectionType = TypeVar('SimpleNetworkConnectionType', bound='SimpleNetworkConnectionProtocol')


class BaseContext(TypedDict):
    protocol_name: str
    address: str
    peer: str
    own: str
    alias: str
    server: str
    client: str


class AFINETContext(BaseContext):
    host: str
    port: int


class SFTPContext(AFINETContext):
    username: str


class AFUNIXContext(BaseContext):
    fd: int


class NamedPipeContext(BaseContext):
    handle: int



from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
from aionetworking.utils import ipv4_or_ipv6
import configparser
import datetime
from pathlib import Path
from typing import Tuple, Type, Dict, Union, Any, TypedDict


def load_cert_file(cert: Path) -> x509.Certificate:
    pem_data = cert.read_bytes()
    return x509.load_pem_x509_certificate(pem_data, default_backend())


def load_cert_expiry_time(cert: Path) -> datetime.datetime:
    cert_data = load_cert_file(cert)
    return cert_data.not_valid_after


def generate_privkey(path: Path = None, public_exponent: int = 65537, key_size: int = 2048, passphrase: str = None) -> \
        rsa.RSAPrivateKeyWithSerialization:
    key = rsa.generate_private_key(
        public_exponent=public_exponent,
        key_size=key_size,
        backend=default_backend()
    )
    if passphrase:
        encryption_algorithm = serialization.BestAvailableEncryption(passphrase.encode())
    else:
        encryption_algorithm = serialization.NoEncryption()
    data = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=encryption_algorithm
    )
    if path:
        path.write_bytes(data)
    return key


class DistinguishedName(TypedDict):
    C: str
    ST: str
    L: str
    O: str
    OU: str
    CN: str
    emailAddress: str


test_distinguished_name: DistinguishedName = DistinguishedName(C='IE', ST='Dublin', L='Dublin', O='AIONetworking',
                                                               OU='testing', CN='localhost',
                                                               emailAddress='aionetworking@example.com')


def get_name_type(key: str) -> Type[x509.GeneralName]:
    if key.startswith('DNS'):
        return x509.DNSName
    elif key.startswith('IP'):
        return x509.IPAddress
    elif key.startswith('RID'):
        return x509.RegisteredID
    elif key.startswith('DN'):
        return x509.DirectoryName
    raise KeyError(f'{key} is not a valid Alternative Subject Name Type')


def generate_cert(path: Path, private_key: rsa.RSAPrivateKey, validity: int = 365, distinguished_name: DistinguishedName = None,
                  basicConstraints: Dict[str, Union[str, bool]] = None, keyUsage: Dict[str, bool] = None,
                  subjectAltName: [str, str] = None) -> x509.Certificate:
    distinguished_name = distinguished_name or test_distinguished_name
    subject = issuer = x509.Name([
            x509.NameAttribute(NameOID.COUNTRY_NAME, distinguished_name['C']),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, distinguished_name['ST']),
            x509.NameAttribute(NameOID.LOCALITY_NAME, distinguished_name['L']),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, distinguished_name['O']),
            x509.NameAttribute(NameOID.ORGANIZATIONAL_UNIT_NAME, distinguished_name['O']),
            x509.NameAttribute(NameOID.COMMON_NAME, distinguished_name['CN']),
            x509.NameAttribute(NameOID.EMAIL_ADDRESS, distinguished_name['emailAddress'])
        ])
    cert = x509.CertificateBuilder().subject_name(
    subject
        ).issuer_name(
    issuer
        ).public_key(
    private_key.public_key()
        ).serial_number(
    x509.random_serial_number()
        ).not_valid_before(
    datetime.datetime.utcnow()
        ).not_valid_after(
        datetime.datetime.utcnow() + datetime.timedelta(days=validity)
        )
    if basicConstraints:
        basicConstraints = {key.lower(): True if value is True or value.lower() == 'true' else False for key, value in basicConstraints.items()}
        for key in ('ca', 'path_length'):
            if key not in basicConstraints:
                basicConstraints[key] = None
        cert.add_extension(
            x509.BasicConstraints(**basicConstraints), critical=True,
        )
    if keyUsage:
        key_usage = {
            'crl_sign': keyUsage.get('crlSign', False),
            'data_encipherment': keyUsage.get('dataEncipherment', False),
            'digital_signature': keyUsage.get('digitalSignature', False),
            'content_commitment': keyUsage.get('nonRepudiation', False),
            'key_agreement': keyUsage.get('keyAgreement', False),
            'key_cert_sign': keyUsage.get('keyCertSign', False),
            'key_encipherment': keyUsage.get('keyEncipherment', False),
            'encipher_only': keyUsage.get('encipherOnly', False),
            'decipher_only': keyUsage.get('decipherOnly', False),
        }
        cert.add_extension(
            x509.KeyUsage(**key_usage), critical=True,
        )
    if subjectAltName:
        alt_names = []
        for key, value in subjectAltName.items():
            if key.startswith('IP'):
                value = ipv4_or_ipv6(value)
            san_type = get_name_type(key)
            alt_names.append(san_type(value))
        cert = cert.add_extension(
            x509.SubjectAlternativeName(alt_names), critical=False,
        )
    cert = cert.sign(private_key, hashes.SHA256(), default_backend())
    data = cert.public_bytes(serialization.Encoding.PEM)
    Path(path).write_bytes(data)
    return cert


def generate_signed_key_cert(base_path: Path, privkey_filename='privkey.pem', cert_filename='cert.pem',
                             public_exponent: int = 65537, key_size: int = 2048, passphrase: str = None,
                             validity: int = 365, distinguished_name: DistinguishedName = None,
                             basicConstraints: Dict[str, bool] = None, keyUsage: Dict[str, bool] = None,
                             subjectAltName: [str, str] = None) -> Tuple[x509.Certificate, Path, rsa.RSAPrivateKey, Path]:
    key_path = Path(base_path / privkey_filename)
    cert_path = Path(base_path / cert_filename)
    key = generate_privkey(key_path, public_exponent=public_exponent, key_size=key_size, passphrase=passphrase)
    cert = generate_cert(cert_path, key, validity=validity, distinguished_name=distinguished_name,
                         basicConstraints=basicConstraints, keyUsage=keyUsage, subjectAltName=subjectAltName)
    return cert, cert_path, key, key_path


def read_from_openssl_conf_file(path: Union[Path, str]) -> Dict[str, Any]:
    config = configparser.ConfigParser()
    config.optionxform = str
    with open(path, 'r') as f:
        config.read_file(open(path))
    distinguished_name_section_name = config.get('req', 'distinguished_name')
    distinguished_name = DistinguishedName(**config[distinguished_name_section_name])
    req_extension_section_name = config.get('req', 'req_extension', fallback=None)
    if req_extension_section_name:
        req_extension_section = config[req_extension_section_name]
        basicConstraints = req_extension_section.get('basicConstraints', fallback=None)
        if basicConstraints:
            basic_constraints = {}
            for basic_constraint in basicConstraints.replace(' ', '').split(','):
                key, value = basic_constraint.split(':')
                basic_constraints[key] = value
        else:
            basic_constraints = None
        keyUsage = req_extension_section.get('keyUsage', fallback=None)
        if keyUsage:
            key_usage = {}
            for key in keyUsage.replace(' ', '').split(','):
                key_usage[key] = True
        else:
            key_usage = None
        subject_alt_name = req_extension_section.get('subjectAltName', '')
        if subject_alt_name.startswith('@'):
            subject_alt_names_section_name = subject_alt_name.replace('@', '')
            subject_alt_names = config[subject_alt_names_section_name]
        elif subject_alt_name:
            subject_alt_names = {}
            for alt_name in subject_alt_name.replace(' ', '').split(','):
                key, value = alt_name.split(':')
                subject_alt_names[key] = value
        else:
            subject_alt_names = None
        return dict(distinguished_name=distinguished_name, basicConstraints=basic_constraints,
                    keyUsage=key_usage, subjectAltName=subject_alt_names)
    return dict(distinguished_name=distinguished_name)


def generate_cert_from_openssl_conf_file(conf_path: Union[str, Path], out_path: Path, key: rsa.RSAPrivateKey,
                                         validity: int = 365):
    kwargs = read_from_openssl_conf_file(conf_path)
    return generate_cert(out_path, key, validity=validity, **kwargs)


def generate_signed_key_cert_from_openssl_conf_file(conf_path: Union[str, Path], out_path: Path,
                                                    privkey_filename='privkey.pem', cert_filename='cert.pem',
                                                    public_exponent: int = 65537, key_size: int = 2048,
                                                    passphrase: str = None, validity: int = 365):
    kwargs = read_from_openssl_conf_file(conf_path)
    return generate_signed_key_cert(out_path, privkey_filename=privkey_filename, cert_filename=cert_filename,
                                    public_exponent=public_exponent, key_size=key_size, passphrase=passphrase,
                                    validity=validity, **kwargs)



import yaml
from .sftp import SFTPServer


def sftp_server_constructor(loader, node) -> SFTPServer:
    value = loader.construct_mapping(node) if node.value else {}
    return SFTPServer(**value)


def load_sftp_server(Loader=yaml.SafeLoader):
    yaml.add_constructor('!SFTPServer', sftp_server_constructor, Loader=Loader)



from .adaptors import ReceiverAdaptor, SenderAdaptor, BaseAdaptorProtocol
from .connections_manager import ConnectionsManager, connections_manager
from .connections import (BaseConnectionProtocol, NetworkConnectionProtocol, TCPServerConnection, TCPClientConnection,
                          BaseStreamConnection, BaseUDPConnection, UDPServerConnection, UDPClientConnection,
                          UDPConnectionMixinProtocol)
from .exceptions import *
from .protocol_factories import (BaseProtocolFactory, BaseDatagramProtocolFactory, StreamClientProtocolFactory,
                                 StreamServerProtocolFactory, DatagramServerProtocolFactory, DatagramClientProtocolFactory)
from .ssl import ServerSideSSL, ClientSideSSL


from datetime import datetime, timedelta, time
import inflect

from typing import Any


class LoggingDatetime:

    def __init__(self, datefmt: str = "%Y-%M-%d %H:%M:%S"):
        self.dt = datetime.now()
        self._datefmt = datefmt

    def __gt__(self, other):
        return self.dt > other.dt

    def __ge__(self, other):
        return self.dt >= other.dt

    def __le__(self, other):
        return self.dt >= other.dt

    def __lt__(self, other):
        return self.dt < other.dt

    def __eq__(self, other):
        return self.dt == other.dt

    def __getattr__(self, item: str) -> Any:
        return getattr(self.dt, item)

    def __str__(self) -> str:
        return self.dt.strftime(self._datefmt)

    @property
    def date(self) -> str:
        return self.dt.strftime("%D")

    @property
    def time(self) -> str:
        return self.dt.strftime("%H:%M:%S")

    @property
    def fulltime(self) -> time:
        return self.dt.time()

    @property
    def timestamp(self) -> float:
        return self.dt.timestamp()


class LoggingTimeDelta:
    def __init__(self, start_time: LoggingDatetime, end_time: LoggingDatetime):
        if start_time and end_time:
            self.td = end_time.dt - start_time.dt
        else:
            self.td = timedelta()
        self._divisor = self.td.total_seconds() or 1.00

    def __int__(self):
        return int(self._divisor)

    def __getattr__(self, item: str) -> Any:
        return getattr(self.td, item)

    def __rtruediv__(self, other: float) -> float:
        return other.__class__(other / self._divisor)

    def __str__(self) -> str:
        return str(self.td)

    def __float__(self) -> float:
        return self._divisor

    @property
    def _minutes(self) -> float:
        return self.td.total_seconds() / 60

    @property
    def _hours(self) -> float:
        return self._minutes / 60

    @property
    def mins(self) -> int:
        return round(self._minutes)

    @property
    def hours(self) -> int:
        return round(self._hours)

    @property
    def timestamp(self) -> str:
        hours = self.td.days * 24 + self.td.seconds // 3600
        minutes = (self.td.seconds % 3600) // 60
        seconds = self.td.seconds % 60
        return "{0}:{1}:{2}".format(hours, minutes, seconds)

    def full_time(self) -> timedelta:
        return self.td


class BytesSizeRate(float):

    def __add__(self, other) -> float:
        return self.__class__(super().__add__(other))

    def __sub__(self, other) -> float:
        return self.__class__(super().__sub__(other))

    def __mul__(self, other) -> float:
        return self.__class__(super().__mul__(other))

    def __floordiv__(self, other) -> float:
        return self.__class__(super().__sub__(other))

    def __truediv__(self, other) -> float:
        return self.__class__(super().__sub__(other))

    def __rtruediv__(self, other) -> float:
        return self.__class__(super().__sub__(float(other)))

    @property
    def bits(self) -> float:
        return float(self) * 8

    @property
    def bytes(self) -> float:
        return float(self)

    @property
    def kb(self) -> float:
        return float(self) / 1024

    @property
    def mb(self) -> float:
        return float(self) / 1024 / 1024

    @property
    def gb(self) -> float:
        return float(self) / 1024 / 1024 / 1024

    @property
    def tb(self) -> float:
        return float(self) / 1024 / 1024 / 1024 / 1024

    def rate(self, interval: float) -> float:
        return self.__class__(self / interval)

    def average(self, num) -> float:
        return self / (num or 1)


class BytesSize(int):

    def __add__(self, other) -> 'BytesSize':
        return BytesSize(other + int(self))

    def __floordiv__(self, other) -> BytesSizeRate:
        return BytesSizeRate(float(self).__floordiv__(float(other)))

    def __truediv__(self, other) -> BytesSizeRate:
        return BytesSizeRate(float(self).__truediv__(float(other)))

    @property
    def bits(self):
        return int(self) * 8

    @property
    def bytes(self) -> int:
        return int(self)

    @property
    def kb(self) -> float:
        return int(self) / 1024

    @property
    def mb(self) -> float:
        return int(self) / 1024 / 1024

    @property
    def gb(self) -> float:
        return int(self) / 1024 / 1024 / 1024

    @property
    def tb(self) -> float:
        return int(self) / 1024 / 1024 / 1024 / 1024

    def __repr__(self):
        return f"{int(self)} bytes"


class MsgsCount:
    sent: int = 0
    received: int = 0
    processed: int = 0
    filtered: int = 0
    failed: int = 0
    first_received = None
    last_received = None
    first_sent = None
    last_sent = None
    last_processed = None

    @property
    def total_done(self) -> int:
        return self.processed + self.filtered + self.failed

    @property
    def percent_filtered(self) -> float:
        return self.filtered / (self.total_done or 1)

    @property
    def percent_failed(self) -> float:
        return self.failed / (self.total_done or 1)

    @property
    def percent_processed(self) -> float:
        return self.processed / (self.total_done or 1)

    @property
    def receive_interval(self) -> LoggingTimeDelta:
        return LoggingTimeDelta(self.first_received, self.last_received)

    @property
    def send_rate(self) -> float:
        return self.sent / self.send_interval

    @property
    def send_interval(self) -> LoggingTimeDelta:
        return LoggingTimeDelta(self.first_sent, self.last_sent)

    @property
    def processing_time(self) -> LoggingTimeDelta:
        return LoggingTimeDelta(self.first_received, self.last_processed)

    @property
    def buffer_receive_rate(self) -> float:
        return self.received / self.receive_interval

    @property
    def buffer_processing_rate(self) -> float:
        return self.received / self.processing_time

    @property
    def processing_rate(self) -> float:
        return self.processed / self.processing_time


p = inflect.engine()



import datetime
from dataclasses import dataclass, field
from pathlib import Path
from pprint import pformat

from aionetworking import settings
from aionetworking.logging.loggers import get_connection_logger_receiver
from aionetworking.types.logging import ConnectionLoggerType
from aionetworking.utils import aone, dataclass_getstate, dataclass_setstate

from .protocols import MessageObject, Codec
from typing import AsyncGenerator, Any, Dict, Sequence, Type, Optional, Protocol
from aionetworking.types.formats import MessageObjectType, CodecType
from aionetworking.types.networking import BaseContext


def current_time() -> datetime.datetime:
    # Required to work with freeze_gun
    return datetime.datetime.now()


@dataclass
class BaseMessageObject(MessageObject, Protocol):
    name = None
    codec_cls = None
    id_attr = 'id'
    stats_logger = None

    encoded: bytes
    decoded: Any = None
    context: Dict[str, Any] = field(default_factory=dict, compare=False, repr=False, hash=False)
    parent_logger: ConnectionLoggerType = field(default_factory=get_connection_logger_receiver, compare=False, hash=False, repr=False)
    system_timestamp: datetime = field(default_factory=current_time, compare=False, repr=False, hash=False)
    received: bool = field(default=True, compare=False, repr=False)

    def __post_init__(self):
        self.logger = self.parent_logger.new_msg_logger(self)

    @property
    def received_or_sent(self) -> str:
        return "RECEIVED" if self.received else "SENT"

    @classmethod
    def _get_codec_kwargs(cls) -> Dict:
        return {}

    @classmethod
    def get_codec(cls, first_buffer_received: Optional[bytes] = None, **kwargs) -> CodecType:
        kwargs.update(cls._get_codec_kwargs())
        return cls.codec_cls(cls, **kwargs)

    @property
    def sender(self) -> str:
        return self.context.get('host', self.context.get('peer'))

    @property
    def receiver(self) -> str:
        return self.full_receiver.split(':')[0]

    @property
    def full_receiver(self) -> str:
        return self.context.get('own')

    @property
    def address(self) -> str:
        return self.context['address']

    @property
    def full_sender(self) -> str:
        return self.context['peer']

    def __getstate__(self):
        return dataclass_getstate(self)

    def __setstate__(self, state):
        dataclass_setstate(self, state)

    def get(self, item, default=None):
        try:
            return self.decoded[item]
        except KeyError:
            return default

    @property
    def uid(self) -> Any:
        try:
            return self.decoded[self.id_attr]
        except (AttributeError, KeyError, TypeError):
            return id(self)

    @property
    def request_id(self) -> Any:
        try:
            return self.decoded.get(self.id_attr)
        except (AttributeError, KeyError):
            return None

    @property
    def pformat(self) -> str:
        return pformat(self.decoded, compact=True, width=120)

    @property
    def timestamp(self) -> datetime.datetime:
        return self.system_timestamp

    def filter(self) -> bool:
        return False

    def __str__(self):
        return f"{self.name} {self.uid}"


@dataclass
class BaseCodec(Codec):
    codec_name = ''
    read_mode = 'rb'
    write_mode = 'wb'
    append_mode = 'ab'
    log_msgs = True
    supports_notifications = False

    msg_obj: Type[MessageObjectType]
    context: BaseContext = field(default_factory=dict)
    logger: ConnectionLoggerType = field(default_factory=get_connection_logger_receiver, compare=False, hash=False, repr=False)

    def __post_init__(self):
        self.context = self.context or {}

    async def decode(self, encoded: bytes, **kwargs) -> AsyncGenerator[Sequence[bytes], None]:
        yield encoded, encoded

    async def encode(self, decoded: Any, **kwargs) -> bytes:
        return decoded

    async def decode_one(self, encoded: bytes, **kwargs) -> Any:
        return await aone(self.decode(encoded, **kwargs))

    async def create_object(self, encoded: bytes, decoded: Any, **kwargs) -> MessageObjectType:
        return self.msg_obj(encoded, decoded, **kwargs)

    async def _from_buffer(self, encoded: bytes, **kwargs) -> AsyncGenerator[MessageObjectType, None]:
        async for encoded, decoded in self.decode(encoded, **kwargs):
            yield await self.create_object(encoded, decoded, parent_logger=self.logger, **kwargs)

    async def decode_buffer(self, encoded: bytes, context: BaseContext = None, source: str = 'buffer',
                            extra_log_context: Dict[str, Any] = None, **kwargs) -> AsyncGenerator[MessageObjectType, None]:
        if context:
            complete_context = self.context.copy()
            complete_context.update(context)
        else:
            complete_context = self.context
        i = 0
        async for msg in self._from_buffer(encoded, context=complete_context, **kwargs):
            if self.log_msgs:
                self.logger.on_msg_decoded(msg, extra=extra_log_context)
            yield msg
            i += 1
        self.logger.on_buffer_decoded(encoded, i, source=source)

    async def encode_obj(self, decoded: Any, **kwargs) -> MessageObjectType:
        try:
            encoded = await self.encode(decoded, **kwargs)
            return await self.create_object(encoded, decoded, context=self.context, received=False,
                                            parent_logger=self.logger, **kwargs)
        except Exception as exc:
            obj = await self.create_object(b'', decoded, context=self.context, parent_logger=self.logger, received=False, **kwargs)
            self.logger.on_encode_failed(obj, exc)

    async def from_file(self, file_path: Path, **kwargs) -> AsyncGenerator[MessageObjectType, None]:
        self.logger.debug('Loading new %s messages from %s', self.codec_name, file_path)
        async with settings.FILE_OPENER(file_path, self.read_mode) as f:
            encoded = await f.read()
        async for item in self.decode_buffer(encoded, source=f"file {file_path}", **kwargs):
            yield item

    async def one_from_file(self, file_path: Path, **kwargs) -> MessageObjectType:
        return await aone(self.from_file(file_path, **kwargs))



from dataclasses import dataclass
from .base import BaseRequester


@dataclass
class EchoRequester(BaseRequester):
    methods = ('echo', 'make_exception')
    notification_methods = ('subscribe', 'simple')
    last_id = 0

    def _make_request(self, method: str, request_id: bool):
        request = {}
        if request_id:
            self.last_id += 1
            request['id'] = self.last_id
        request['method'] = method
        return request

    def echo(self):
        return self._make_request('echo', True)

    def simple(self):
        return self._make_request('simple', True)

    def make_exception(self):
        return self._make_request('echo_typo', True)

    def subscribe(self):
        return self._make_request('send_notification', False)



import asyncio
from abc import abstractmethod
import datetime
from dataclasses import dataclass, field
from pathlib import Path

from aionetworking.types.actions import ActionType
from aionetworking.types.requesters import RequesterType
from aionetworking.types.logging import LoggerType
from aionetworking.types.formats import MessageObjectType
from aionetworking.utils import inherit_on_type_checking_only

from aionetworking.types.networking import AdaptorType
from .transports import TransportType

from typing import Any, AsyncGenerator, Generator, Optional, Sequence, Union, Dict, Tuple, Type, Protocol


class ProtocolFactoryProtocol(Protocol):

    @abstractmethod
    def __call__(self) -> Union['ProtocolFactoryProtocol', 'NetworkConnectionProtocol']: ...

    @abstractmethod
    def is_owner(self, connection: 'NetworkConnectionProtocol') -> bool: ...

    @abstractmethod
    def set_name(self, name: str, peer_prefix: str): ...

    @abstractmethod
    async def wait_all_closed(self) -> None: ...

    @abstractmethod
    async def wait_num_has_connected(self, num: int) -> None: ...

    @abstractmethod
    async def close_actions(self) -> None: ...

    @abstractmethod
    async def start(self, logger: LoggerType = None) -> None: ...

    @abstractmethod
    async def close(self) -> None: ...


@dataclass
class ConnectionProtocol(Protocol):

    @property
    @abstractmethod
    def peer(self) -> str: ...

    @abstractmethod
    def is_child(self, parent_name: str) -> bool: ...

    @abstractmethod
    def finish_connection(self, exc: Optional[BaseException]) -> None: ...

    @abstractmethod
    async def wait_closed(self) -> None: ...

    @abstractmethod
    async def wait_connected(self) -> None: ...

    @abstractmethod
    def is_connected(self) -> bool: ...

    @abstractmethod
    def send(self, data: bytes) -> Optional[asyncio.Future]: ...


@dataclass
class ConnectionDataclassProtocol(ConnectionProtocol, Protocol):
    name = None

    parent_name: str = None
    action: ActionType = None
    preaction: ActionType = None
    requester: RequesterType = None
    dataformat: Type[MessageObjectType] = None
    context: Dict[str, Any] = field(default_factory=dict, metadata={'pickle': True})
    peer_prefix: str = ''
    last_msg: datetime.datetime = field(default=None, init=False, compare=False, hash=False)
    timeout: Union[int, float] = None

    adaptor_cls: Type[AdaptorType] = field(default=None, init=False)
    _adaptor: AdaptorType = field(default=None, init=False)


@dataclass
class NetworkConnectionProtocol(ConnectionDataclassProtocol, Protocol):
    @abstractmethod
    def initialize_connection(self, transport: TransportType, peer: Tuple[str, int] = None) -> bool: ...


@dataclass
class SimpleNetworkConnectionProtocol(Protocol):
    peer: str
    parent_name: str
    queue: asyncio.Queue

    @abstractmethod
    def encode_and_send_msg(self, msg_decoded: Any) -> None: ...


class UDPConnectionMixinProtocol(Protocol): ...


class UDPConnectionProtocol(UDPConnectionMixinProtocol, NetworkConnectionProtocol, Protocol):  ...


class BaseAdaptorProtocol(Protocol):

    @abstractmethod
    def send_data(self, msg_encoded: bytes) -> asyncio.Future: ...

    @abstractmethod
    def encode_and_send_msg(self, msg_decoded: Any) -> None: ...

    @abstractmethod
    def encode_and_send_msgs(self, decoded_msgs: Sequence[Any]) -> None: ...

    @abstractmethod
    def on_data_received(self, buffer: bytes, timestamp: datetime.datetime = None) -> asyncio.Future: ...


class AdaptorProtocol(BaseAdaptorProtocol, Protocol):

    @abstractmethod
    async def close(self, exc: Optional[BaseException] = None) -> None: ...


@inherit_on_type_checking_only
class AdaptorProtocolGetattr(BaseAdaptorProtocol, Protocol):

    def send_data(self, msg_encoded: bytes) -> None: ...

    def encode_and_send_msg(self, msg_decoded: Any) -> None: ...

    def encode_and_send_msgs(self, decoded_msgs: Sequence[Any]) -> None: ...

    def on_data_received(self, buffer: bytes, timestamp: datetime.datetime = None) -> asyncio.Future: ...


class SenderAdaptorMixinProtocol(Protocol):
    @abstractmethod
    async def wait_notification(self) -> MessageObjectType: ...

    @abstractmethod
    def get_notification(self) -> MessageObjectType: ...

    @abstractmethod
    async def wait_notifications(self) -> AsyncGenerator[MessageObjectType, None]:
        yield

    @abstractmethod
    def all_notifications(self) -> Generator[MessageObjectType, None, None]: ...

    @abstractmethod
    async def send_data_and_wait(self, request_id: Any, encoded: bytes) -> asyncio.Future: ...

    @abstractmethod
    async def send_msg_and_wait(self, msg_obj: MessageObjectType) -> asyncio.Future: ...

    @abstractmethod
    async def encode_send_wait(self, decoded: Any) -> asyncio.Future: ...

    @abstractmethod
    async def play_recording(self, file_path: Path, hosts: Sequence = (), timing: bool = True) -> None: ...


class SenderAdaptorProtocol(ConnectionProtocol, Protocol): ...


@inherit_on_type_checking_only
class SenderAdaptorGetattr(SenderAdaptorMixinProtocol, Protocol):
    async def wait_notification(self) -> MessageObjectType: ...

    def get_notification(self) -> MessageObjectType: ...

    async def wait_notifications(self) -> AsyncGenerator[MessageObjectType, None]:
        yield

    def all_notifications(self) -> Generator[MessageObjectType, None, None]: ...

    async def send_data_and_wait(self, request_id: Any, encoded: bytes) -> asyncio.Future: ...

    async def send_msg_and_wait(self, msg_obj: MessageObjectType) -> asyncio.Future: ...

    async def encode_send_wait(self, decoded: Any) -> asyncio.Future: ...

    async def play_recording(self, file_path: Path, hosts: Sequence = (), timing: bool = True) -> None: ...



from abc import abstractmethod
from aionetworking.types.networking import ConnectionType
from typing import Protocol


class SenderProtocol(Protocol):

    @property
    @abstractmethod
    def dst(self) -> str: ...

    @abstractmethod
    async def __aenter__(self) -> ConnectionType: ...

    @abstractmethod
    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None: ...

    @abstractmethod
    def is_started(self) -> bool: ...

    @abstractmethod
    def is_closing(self) -> bool: ...

    @abstractmethod
    async def connect(self) -> ConnectionType: ...

    @abstractmethod
    async def close(self): ...


from asyncio import transports
from typing import Tuple, Any, Union, Optional


class DatagramTransportWrapper:
    def __init__(self, transport: transports.DatagramTransport, peer: Optional[Tuple[str, int]] = None):
        self._transport = transport
        self._peer = peer

    def __getattr__(self, name):
        return getattr(self._transport, name)

    def get_extra_info(self, name: Any, default: Any = ...) -> Any:
        if name == 'peername' and self._peer:
            return self._peer
        return self._transport.get_extra_info(name, default=default)

    def write(self, data: Any) -> None:
        if self._peer:
            self._transport.sendto(data, self._peer)
        else:
            self._transport.sendto(data)


TransportType = Union[transports.Transport, DatagramTransportWrapper]



from .servers import TCPServer, UDPServer, UnixSocketServer
import yaml


def tcp_server_constructor(loader, node) -> TCPServer:
    value = loader.construct_mapping(node) if node.value else {}
    return TCPServer(**value)


def load_tcp_server(Loader=yaml.SafeLoader):
    yaml.add_constructor('!TCPServer', tcp_server_constructor, Loader=Loader)


def udp_server_constructor(loader, node) -> UDPServer:
    value = loader.construct_mapping(node) if node.value else {}
    return UDPServer(**value)


def load_udp_server(Loader=yaml.SafeLoader):
    yaml.add_constructor('!UDPServer', udp_server_constructor, Loader=Loader)


def pipe_server_constructor(loader, node) -> UnixSocketServer:
    value = loader.construct_mapping(node) if node.value else {}
    return UnixSocketServer(**value)


def load_pipe_server(Loader=yaml.SafeLoader):
    yaml.add_constructor('!PipeServer', pipe_server_constructor, Loader=Loader)



import asyncio
from dataclasses import dataclass, field, replace
import datetime

from aionetworking.actions.protocols import ActionProtocol
from aionetworking.formats.base import BaseMessageObject
from aionetworking.futures import TaskScheduler
from aionetworking.types.requesters import RequesterType
from aionetworking.logging.loggers import get_logger_receiver
from aionetworking.logging.utils_logging import p
from aionetworking.types.logging import LoggerType
from aionetworking.types.networking import BaseContext
from aionetworking.utils import dataclass_getstate, dataclass_setstate, addr_tuple_to_str, IPNetwork
from .transports import DatagramTransportWrapper


from .connections_manager import connections_manager
from .connections import TCPClientConnection, TCPServerConnection, UDPServerConnection, UDPClientConnection
from .protocols import ProtocolFactoryProtocol
from aionetworking.types.networking import ProtocolFactoryType,  NetworkConnectionType

from typing import Optional, Tuple, Type, Union, Sequence, Dict, Any


@dataclass
class BaseProtocolFactory(ProtocolFactoryProtocol):
    full_name = ''
    peer_prefix = ''
    connection_cls: Type[NetworkConnectionType] = field(default=None, init=False)
    action: ActionProtocol = None
    preaction: ActionProtocol = None
    requester: RequesterType = None
    dataformat: Type[BaseMessageObject] = None
    logger: LoggerType = field(default_factory=get_logger_receiver)
    pause_reading_on_buffer_size: int = None
    hostname_lookup: bool = False
    expire_connections_after_inactive_minutes: Union[int, float] = 0
    expire_connections_check_interval_minutes: Union[int, float] = 1
    aliases: Dict[str, str] = field(default_factory=dict)
    allowed_senders: Sequence[IPNetwork] = field(default_factory=tuple)
    check_peer_cert_expiry: int = 7
    codec_config: Dict[str, Any] = field(default_factory=dict, metadata={'pickle': True})
    timeout: int = None
    quiet: bool = False
    _scheduler: TaskScheduler = field(default_factory=TaskScheduler, init=False)
    context: BaseContext = field(default_factory=dict, init=False, compare=False, repr=False)

    def __post_init__(self):
        if self.preaction:
            self.preaction = replace(self.preaction, quiet=True)
        if self.action:
            self.action = replace(self.action, quiet=self.quiet)
        if self.requester:
            self.requester = replace(self.requester)

    async def start(self, context: BaseContext = None, logger: LoggerType = None) -> None:
        self.context = context or self.context
        self.logger = logger or self.logger
        coros = []
        if self.action:
            coros.append(self.action.start(logger=logger))
        if self.preaction:
            coros.append(self.preaction.start(logger=logger))
        if self.requester:
            coros.append(self.requester.start(logger=logger))
        await asyncio.gather(*coros)
        if self.expire_connections_after_inactive_minutes:
            self.logger.info('Connections will expire after %s',
                             p.no('minute', self.expire_connections_check_interval_minutes))
            self._scheduler.call_cb_periodic(self.expire_connections_check_interval_minutes * 60,
                                             self.check_expired_connections,
                                             task_name=f'Check expired connections for {self.full_name}')

    def __call__(self) -> NetworkConnectionType:
        return self._new_connection()

    def _additional_connection_kwargs(self) -> Dict[str, Any]:
        return {}

    def _new_connection(self) -> NetworkConnectionType:
        self.logger.debug('Creating new connection')
        return self.connection_cls(parent_name=self.full_name, peer_prefix=self.peer_prefix, action=self.action,
                                   preaction=self.preaction, requester=self.requester, dataformat=self.dataformat,
                                   pause_reading_on_buffer_size=self.pause_reading_on_buffer_size, logger=self.logger,
                                   hostname_lookup=self.hostname_lookup, allowed_senders=self.allowed_senders,
                                   context=self.context.copy(), check_peer_cert_expiry=self.check_peer_cert_expiry,
                                   timeout=self.timeout, codec_config=self.codec_config,
                                   **self._additional_connection_kwargs())

    def __getstate__(self):
        return dataclass_getstate(self)

    def __setstate__(self, state):
        dataclass_setstate(self, state)

    def set_name(self, full_name: str, peer_prefix: str) -> None:
        self.full_name = full_name
        self.peer_prefix = peer_prefix

    def is_owner(self, connection: NetworkConnectionType) -> bool:
        return connection.is_child(self.full_name)

    @property
    def num_connections(self) -> int:
        return connections_manager.num_connections(self.full_name)

    async def wait_num_has_connected(self, num: int) -> None:
        await connections_manager.wait_num_has_connected(self.full_name, num)

    async def wait_num_connected(self, num: int) -> None:
        await connections_manager.wait_num_connections(self.full_name, num)

    async def wait_all_closed(self) -> None:
        await connections_manager.wait_num_connections(self.full_name, 0)

    async def close_actions(self) -> None:
        coros = []
        if self.action:
            coros.append(self.action.close())
        if self.preaction:
            coros.append(self.preaction.close())
        if self.requester:
            coros.append(self.requester.close())
        await asyncio.gather(*coros)

    @staticmethod
    def close_connection(conn: NetworkConnectionType, exc: Optional[Exception]):
        conn.close()

    def close_all_connections(self, exc: Optional[Exception]) -> None:
        for conn in filter(self.is_owner, list(connections_manager)):
            self.close_connection(conn, exc)

    def check_expired_connections(self):
        now = datetime.datetime.now()
        for conn in filter(self.is_owner, list(connections_manager)):
            if (now - conn.last_msg).total_seconds() >= (self.expire_connections_after_inactive_minutes * 60):
                self.close_connection(conn, None)

    async def close(self) -> None:
        num_connections = self.num_connections
        if num_connections:
            self.logger.warning('Waiting to complete tasks for %s', p.no('connection', num_connections))
        await asyncio.wait([asyncio.create_task(self._scheduler.close()), asyncio.create_task(self.wait_num_connected(0))], timeout=self.timeout)
        self.logger.info('Protocol factory tasks finished, and all connections have been closed')
        await asyncio.wait_for(self.close_actions(), self.timeout)
        self.logger.info('Actions complete')
        connections_manager.clear_server(self.full_name)


@dataclass
class StreamServerProtocolFactory(BaseProtocolFactory):
    connection_cls = TCPServerConnection


@dataclass
class StreamClientProtocolFactory(BaseProtocolFactory):
    connection_cls = TCPClientConnection


@dataclass
class BaseDatagramProtocolFactory(asyncio.DatagramProtocol, BaseProtocolFactory):
    connection_cls: NetworkConnectionType = UDPServerConnection
    transport = None
    sock = None

    def __call__(self: ProtocolFactoryType) -> ProtocolFactoryType:
        return self

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:
        self.transport = transport
        self.sock = self.transport.get_extra_info('sockname')[0:2]

    @staticmethod
    def close_connection(conn: NetworkConnectionType, exc: Optional[Exception]):
        conn.connection_lost(exc)

    def connection_lost(self, exc: Optional[Exception]) -> None:
        self.logger.manage_error(exc)
        self.close_all_connections(exc)

    def error_received(self, exc: Optional[Exception]) -> None:
        self.logger.manage_error(exc)

    def new_peer(self, addr: Tuple[str, int] = None) -> NetworkConnectionType:
        conn = self._new_connection()
        transport = DatagramTransportWrapper(self.transport, addr)
        ok = conn.connection_made(transport)
        if ok:
            return conn

    def datagram_received(self, data: bytes, addr: Tuple[str, int]) -> None:
        addr = addr[0:2]
        peer = self.connection_cls.get_peername(self.peer_prefix, addr_tuple_to_str(addr), addr_tuple_to_str(self.sock))
        conn = connections_manager.get(peer, None)
        if conn:
            conn.data_received(data)
        else:
            conn = self.new_peer(addr)
            if conn:
                conn.data_received(data)

    async def close(self) -> None:
        self.close_all_connections(None)
        await super().close()


@dataclass
class DatagramServerProtocolFactory(BaseDatagramProtocolFactory):
    connection_cls: NetworkConnectionType = UDPServerConnection


@dataclass
class DatagramClientProtocolFactory(BaseDatagramProtocolFactory):
    connection_cls: NetworkConnectionType = UDPClientConnection



from abc import abstractmethod
from dataclasses import dataclass

from aionetworking.types.formats import MessageObjectType
from aionetworking.types.logging import LoggerType

from typing import AsyncGenerator, Any, TypeVar, Protocol


ActionType = TypeVar('ActionType', bound='BaseActionProtocol')


@dataclass
class ActionProtocol(Protocol):
    supports_notifications = False
    task_timeout: int = 10
    quiet: bool = False

    @abstractmethod
    def filter(self, msg: MessageObjectType) -> bool: ...

    @abstractmethod
    async def do_one(self, msg: MessageObjectType) -> Any: ...

    @abstractmethod
    def on_decode_error(self, data: bytes, exc: BaseException) -> Any: ...

    @abstractmethod
    def on_exception(self, msg: MessageObjectType, exc: BaseException) -> Any: ...

    @abstractmethod
    async def get_notifications(self, peer: str) -> AsyncGenerator[None, None]:
        yield

    @abstractmethod
    async def start(self, logger: LoggerType = None) -> None: ...

    @abstractmethod
    async def close(self) -> None: ...




import asyncio
from aionetworking.runners import run_forever
from aionetworking.utils import set_loop_policy
import pywintypes
import win32serviceutil
import win32service
import win32event
import servicemanager
import socket


class BaseAIONetworkingYamlService(win32serviceutil.ServiceFramework):
    _svc_name_ = "AIONetworkingService"
    _svc_display_name_ = "AIONetworking Service"
    _svc_description_ = "AIONetworking Based Service"
    loop = 'proactor'
    conf_file = None
    paths = None

    def __init__(self, args):
        super().__init__(args)
        self.task = None
        socket.setdefaulttimeout(60)

    def SvcStop(self):
        self.loop.call_soon_threadsafe(self.task.cancel)

    def SvcDoRun(self):
        asyncio.run(self.main())

    async def main(self):
        self.loop = asyncio.get_running_loop()
        self.task = asyncio.create_task(run_forever(self.conf_file, paths=self.paths))
        try:
            await self.task
        except asyncio.CancelledError:
            pass





import asyncssh
import asyncio
from dataclasses import dataclass, field, InitVar
from functools import partial

from aionetworking import settings
from aionetworking.networking.sftp import SFTPServerProtocolFactory, SFTPFactory
from .base import BaseNetworkServer

from typing import Dict, Any, Union
from pathlib import Path


@dataclass
class SFTPServer(BaseNetworkServer):
    protocol_factory: SFTPServerProtocolFactory = None
    sftp_factory = SFTPFactory
    name = 'SFTP Server'
    peer_prefix = 'sftp'
    reuse_address: bool = True
    sftp_log_level: InitVar[int] = 1
    allow_scp: bool = False
    server_host_key: Path = ()
    passphrase: str = None
    extra_sftp_kwargs: Dict[str, Any] = field(default_factory=dict)
    base_upload_dir: Path = settings.TEMPDIR / "sftp_received"

    def __post_init__(self, sftp_log_level) -> None:
        super().__post_init__()
        asyncssh.logging.set_debug_level(sftp_log_level)

    @property
    def sftp_kwargs(self) -> Dict[str, Any]:
        kwargs = {
            'allow_scp': self.allow_scp,
            'server_host_keys': [self.server_host_key],
            'passphrase': self.passphrase,
        }
        kwargs.update(self.extra_sftp_kwargs)
        return kwargs

    async def _get_server(self) -> asyncio.AbstractServer:
        return await asyncssh.create_server(self.protocol_factory, self.host, self.port, backlog=self.backlog,
                                            reuse_address=self.reuse_address, reuse_port=self.reuse_port, line_editor=False,
                                            sftp_factory=partial(self.sftp_factory, base_upload_dir=self.base_upload_dir),
                                            **self.sftp_kwargs)


def generate_key_in_path(path: Union[Path, str], alg_name='ssh-rsa'):
    skey = asyncssh.generate_private_key(alg_name)
    skey.write_private_key(str(path))




from abc import abstractmethod
import asyncio
from typing import Protocol


class ReceiverProtocol(Protocol):

    @abstractmethod
    async def wait_started(self): ...

    @abstractmethod
    async def wait_stopped(self): ...

    @abstractmethod
    async def start(self): ...

    @abstractmethod
    async def serve_until_close_signal(self, stop_event: asyncio.Event = None,
                                       restart_event: asyncio.Event = None, notify_pid: int = None) -> None: ...

    @abstractmethod
    async def serve_forever(self) -> None: ...
    
    @abstractmethod
    async def close(self): ...

    @abstractmethod
    def is_started(self) -> bool: ...

    @abstractmethod
    def is_closing(self) -> bool: ...

    @abstractmethod
    async def wait_all_tasks_done(self) -> None: ...



import asyncio
from datetime import datetime, timedelta
from dataclasses import dataclass, field
from typing import Any, Callable, Awaitable, List, Union, Dict, Optional, Type, Coroutine

from aionetworking.utils import set_task_name
from .counters import Counter


@dataclass
class TaskScheduler:
    _counter: Counter = field(default_factory=Counter, init=False)
    _futures: Dict[Any, asyncio.Future] = field(default_factory=dict, init=False)
    _periodic_tasks: List[asyncio.Future] = field(default_factory=list, init=False)
    _current_tasks: List[asyncio.Future] = field(default_factory=list, init=False)

    def task_done(self, future: asyncio.Future) -> None:
        self._counter.decrement()
        if future in self._current_tasks:
            self._current_tasks.remove(future)

    def create_task(self, coro: Coroutine, name: str = None, include_hierarchy: bool = True,
                    separator: str = ':', continuous: bool = False) -> asyncio.Future:
        self._counter.increment()
        task = asyncio.create_task(coro)
        set_task_name(task, name, include_hierarchy=include_hierarchy, separator=separator)
        if not continuous:
            self._current_tasks.append(task)
        return task

    def task_with_callback(self, coro: Awaitable, callback: Callable = None, name: str = None,
                           include_hierarchy: bool = True, separator: str = ':', continuous: bool = False) -> asyncio.Future:
        task = self.create_task(coro, name=name, include_hierarchy=include_hierarchy, separator=separator, continuous=continuous)
        callback = callback or self.task_done
        task.add_done_callback(callback)
        return task

    def create_future(self, name: Any) -> asyncio.Future:
        self._counter.increment()
        fut = asyncio.Future()
        self._futures[name] = fut
        return fut

    def _call_cb_with_task_done(self, callback, *args, **kwargs):
        try:
            callback(*args, **kwargs)
        finally:
            self._counter.decrement()

    def call_soon(self, callback: Callable, *args):
        self._counter.increment()
        asyncio.get_running_loop().call_soon(self._call_cb_with_task_done, callback, *args)

    def future_done(self, name: Any) -> None:
        fut = self._futures.pop(name)
        self.task_done(fut)

    async def run_wait_fut(self, name: Any, callback: Callable, *args, **kwargs) -> Any:
        fut = self.create_future(name)
        try:
            callback(*args, **kwargs)
            await fut
            return fut.result()
        finally:
            self.future_done(name)

    def set_result(self, name: Any, result: Any) -> None:
        self._futures[name].set_result(result)

    def set_exception(self, name: Any, exception: BaseException) -> None:
        self._futures[name].set_exception(exception)

    def cancel_all_futures(self, exc_class: Optional[Type[BaseException]]):
        for name, fut in self._futures.items():
            if not fut.cancelled() and not fut.done():
                fut.set_exception(exc_class())

    @property
    def task_count(self) -> int:
        return self._counter.num

    async def join(self) -> None:
        await self._counter.wait_for(0)

    async def wait_current_tasks(self) -> None:
        if self._current_tasks:
            current_tasks = self._current_tasks.copy()
            await asyncio.wait(current_tasks)

    async def close(self) -> None:
        self.close_nowait()
        await self.join()
        await asyncio.gather(*self._periodic_tasks, return_exceptions=True)
        self._periodic_tasks.clear()

    def close_nowait(self):
        for task in self._periodic_tasks:
            task.cancel()

    @staticmethod
    def get_next_time(delay: Union[int, float]) -> float:
        now = datetime.now()
        hour = now.hour if delay <= 60 else 0
        start_time = datetime(now.year, now.month, now.day, hour, 0, 0)
        while start_time < now + timedelta(minutes=1):
            start_time += timedelta(seconds=delay)
        td = (start_time - now).total_seconds()
        return td

    def get_start_interval(self, fixed_start_time: bool, immediate: bool, delay: Union[int, float]):
        if fixed_start_time:
            return self.get_next_time(delay)
        elif immediate:
            return 0
        else:
            return delay

    @staticmethod
    async def _call_coro_periodic(interval: Union[int, float], async_callback: Callable, *args,
                                  start_time_interval: Union[int, float] = 0, **kwargs):
        # Asyncio.sleep wrong time on MacOS
        await asyncio.sleep(start_time_interval)
        while True:
            coro = async_callback(*args, **kwargs)
            await coro
            await asyncio.sleep(interval)

    def call_coro_periodic(self, interval: Union[int, float], async_callback: Callable, *args,
                           fixed_start_time: bool = False, immediate: bool = False, task_name: str = None,
                           **kwargs) -> None:
        start_time_interval = self.get_start_interval(fixed_start_time, immediate, interval)
        task = asyncio.create_task(
            self._call_coro_periodic(interval, async_callback, start_time_interval=start_time_interval, *args, **kwargs))
        set_task_name(task, task_name)
        self._periodic_tasks.append(task)

    @staticmethod
    async def _call_cb(callback: Callable, *args, **kwargs):
        return callback(*args, **kwargs)

    def call_cb_periodic(self, interval: Union[int, float], cb: Callable, *args, task_name: str = None, **kwargs):
        self.call_coro_periodic(interval, self._call_cb, cb, *args, task_name=task_name, **kwargs)



import yaml
from .base import EmptyAction
from .echo import EchoAction
from .file_storage import FileStorage, BufferedFileStorage


def echo_action_constructor(loader, node) -> EchoAction:
    value = loader.construct_mapping(node) if node.value else {}
    return EchoAction(**value)


def empty_action_constructor(loader, node) -> EmptyAction:
    value = loader.construct_mapping(node) if node.value else {}
    return EmptyAction(**value)


def file_storage_constructor(loader, node) -> FileStorage:
    value = loader.construct_mapping(node) if node.value else {}
    return FileStorage(**value)


def buffered_file_storage_constructor(loader, node) -> BufferedFileStorage:
    value = loader.construct_mapping(node) if node.value else {}
    return BufferedFileStorage(**value)


def load_echo_action(Loader=yaml.SafeLoader):
    yaml.add_constructor('!EchoAction', echo_action_constructor, Loader=Loader)


def load_empty_action(Loader=yaml.SafeLoader):
    yaml.add_constructor('!EmptyAction', empty_action_constructor, Loader=Loader)


def load_file_storage(Loader=yaml.SafeLoader):
    yaml.add_constructor('!FileStorage', file_storage_constructor, Loader=Loader)


def load_buffered_file_storage(Loader=yaml.SafeLoader):
    yaml.add_constructor('!BufferedFileStorage', buffered_file_storage_constructor, Loader=Loader)





from dataclasses import dataclass, field
from typing import Dict, Any


from aionetworking.types.networking import SimpleNetworkConnectionType
from aionetworking.futures.counters import Counters


endpoint_names = {}


def get_unique_name(name: str) -> str:
    index = endpoint_names.get(name)
    if index:
        full_name = f'{name}_{index}'
        endpoint_names[name] += 1
    else:
        full_name = name
        endpoint_names[name] = 2
    return full_name


def clear_unique_names() -> None:
    endpoint_names.clear()


@dataclass
class ConnectionsManager:
    _connections: Dict[str, SimpleNetworkConnectionType] = field(init=False, default_factory=dict)
    _counters: Counters = field(init=False, default_factory=Counters)

    def clear(self):
        self._connections.clear()
        self._counters.clear()

    def clear_server(self, parent_name: str):
        self._counters.remove(parent_name)

    def add_connection(self, connection: SimpleNetworkConnectionType) -> int:
        self._connections[connection.peer] = connection
        return self._counters.increment(connection.parent_name)

    def remove_connection(self, connection: Any):
        self._connections.pop(connection.peer)

    def decrement(self, connection) -> int:
        return self._counters.decrement(connection.parent_name)

    @property
    def total(self) -> int:
        return len(self._connections)

    def num_connections(self, parent_name: str) -> int:
        return self._counters.get_num(parent_name)

    async def wait_num_connections(self, parent_name: str, num: int) -> None:
        await self._counters.wait_for(parent_name, num)

    async def wait_num_has_connected(self, parent_name: str, num: int) -> None:
        await self._counters.wait_for_total_increments(parent_name, num)

    def get(self, key: str, default: bool = "_raise") -> SimpleNetworkConnectionType:
        if default == "_raise":
            return self._connections[key]
        return self._connections.get(key, default)

    def __iter__(self) -> None:
        for conn in self._connections.values():
            yield conn


connections_manager = ConnectionsManager()



from .protocols import RequesterProtocol
from aionetworking.types.logging import LoggerType
from typing import Protocol


class BaseRequester(RequesterProtocol, Protocol):
    async def start(self, logger: LoggerType = None):
        parent_logger = logger or self.logger
        self.logger = parent_logger.get_child(name='requester')

    async def close(self): ...




import asyncio
import os
from aionetworking.compatibility_os import run
from .conf.yaml_config import SignalServerManager, load_all_tags, server_from_config_file
from typing import Union, Dict
from pathlib import Path


async def run_until_signal(conf: Union[str, Path], paths: Dict[str, Union[str, Path]] = None, notify_pid: int = None,
                           duration: int = None):
    manager = SignalServerManager(conf, paths=paths, notify_pid=notify_pid)
    try:
        await asyncio.wait_for(manager.serve_until_stopped(), timeout=duration)
    except asyncio.TimeoutError:
        pass


async def run_forever(conf, paths: Dict[str, Union[str, Path]] = None, duration: int = None):
    server = server_from_config_file(conf, paths=paths)
    try:
        await asyncio.wait_for(server.serve_forever(), timeout=duration)
    except asyncio.TimeoutError:
        pass


def run_server(conf_file, paths: Dict[str, Union[str, Path]] = None, asyncio_debug: bool = False,
               notify_pid: int = None, duration: int = None):
    debug = asyncio_debug
    if os.name == 'posix':
        run(run_until_signal(conf_file, paths=paths, notify_pid=notify_pid, duration=duration),
            debug=asyncio_debug)
    else:
        try:
            run(run_forever(conf_file, paths=paths, duration=duration), debug=debug)
        except (KeyboardInterrupt, asyncio.TimeoutError):
            pass


def run_server_default_tags(conf_file, paths: Dict[str, Union[str, Path]] = None, asyncio_debug: bool = False,
                            notify_pid: int = None, duration: int = None):
    load_all_tags()
    run_server(conf_file, paths=paths, asyncio_debug=asyncio_debug, notify_pid=notify_pid, duration=duration)
